"""Template jinja filters."""

from datetime import datetime
import copy
import hashlib
import json as json_lib
import random
import re
import jinja2
import slugify
from babel import dates as babel_dates
from babel import numbers as babel_numbers
from grow.common import json_encoder
from grow.common import urls
from grow.common import utils
from grow.templates.tags import _gettext_alias


def _deep_gettext(ctx, fields):
    if isinstance(fields, dict):
        new_dct = {}
        for key, val in fields.items():
            if isinstance(val, (dict, list, set)):
                new_dct[key] = _deep_gettext(ctx, val)
            elif isinstance(val, str):
                new_dct[key] = _gettext_alias(ctx, val)
            else:
                new_dct[key] = val
        return new_dct
    elif isinstance(fields, (list, set)):
        for i, val in enumerate(fields):
            if isinstance(val, (dict, list, set)):
                fields[i] = _deep_gettext(ctx, val)
            elif isinstance(val, str):
                fields[i] = _gettext_alias(ctx, val)
            else:
                fields[i] = val
    return fields


@jinja2.contextfilter
def deeptrans(ctx, obj):
    """Deep translate an object."""
    # Avoid issues (related to sharing the same object across locales and
    # leaking translations from one locale to another) by copying the object
    # before it's sent to deeptrans.
    new_item = copy.deepcopy(obj)
    return _deep_gettext(ctx, new_item)


@jinja2.contextfilter
def expand_partial(_ctx, partial_name):
    """Filter for expanding partial path from name of partial."""
    return '/partials/{0}/{0}.html'.format(partial_name)


@jinja2.contextfilter
def hash_value(_ctx, value, algorithm='sha'):
    """Hash the value using the algorithm."""
    value = value.encode('utf-8')

    if algorithm in ('md5',):
        return hashlib.md5(value).hexdigest()
    if algorithm in ('sha1',):
        return hashlib.sha1(value).hexdigest()
    if algorithm in ('sha224',):
        return hashlib.sha224(value).hexdigest()
    if algorithm in ('sha384',):
        return hashlib.sha384(value).hexdigest()
    if algorithm in ('sha512',):
        return hashlib.sha512(value).hexdigest()
    return hashlib.sha256(value).hexdigest()


@jinja2.contextfilter
def jsonify(_ctx, obj, *args, **kwargs):
    """Filter for JSON dumping an object."""
    return json_lib.dumps(obj, cls=json_encoder.GrowJSONEncoder, *args, **kwargs)


@jinja2.contextfilter
def markdown_filter(ctx, value):
    """Filters content through a markdown processor."""
    doc = ctx['doc']
    m_down = doc.pod.markdown
    try:
        if isinstance(value, str):
            value = value
        value = value or ''
        return m_down.convert(value)
    except UnicodeEncodeError:
        return m_down.convert(value)


@jinja2.contextfilter
def parsedatetime_filter(_ctx, date_string, string_format):
    """Filter dor parsing a datetime."""
    return datetime.strptime(date_string, string_format)


@jinja2.contextfilter
def relative_filter(ctx, path):
    """Calculates the relative path from the current url to the given url."""
    doc = ctx['doc']
    return urls.Url.create_relative_path(
        path, relative_to=doc.url.path)


@jinja2.contextfilter
def render_filter(ctx, template):
    """Creates jinja template from string and renders."""
    if isinstance(template, str):
        template = ctx.environment.from_string(template)
    return template.render(ctx)


@jinja2.contextfilter
def shuffle_filter(_ctx, seq):
    """Shuffles the list into a random order."""
    try:
        result = list(seq)
        random.shuffle(result)
        return result
    except TypeError:
        return seq

def regex_replace():
    """A regex replace filter with regex cache."""
    regex_cache = {}

    def regex_replace_filter(string, find, replace):
        """A template regex filter"""
        if find not in regex_cache:
            regex_cache[find] = re.compile(find)
        return regex_cache[find].sub(replace, string)

    return regex_replace_filter


def slug_filter(pod=None):
    """Filters string to remove url unfriendly characters."""
    use_legacy_slugify = pod and pod.is_enabled(pod.FEATURE_OLD_SLUGIFY)
    def _slug_filter(value, delimiter='-'):
        if not value:
            return value
        if use_legacy_slugify:
            return utils.slugify(value, delimiter)
        return slugify.slugify(value, separator=delimiter)
    return _slug_filter


def wrap_locale_context(func):
    """Wraps the func with the current locale."""
    @jinja2.contextfilter
    def _locale_filter(ctx, value, *args, **kwargs):
        doc = ctx['doc']
        if not kwargs.get('locale', None):
            kwargs['locale'] = str(doc.locale)
        return func(value, *args, **kwargs)
    return _locale_filter


def create_builtin_filters(env, pod=None, locale=None):
    """Filters standard for the template rendering."""
    return (
        ('currency', wrap_locale_context(babel_numbers.format_currency)),
        ('date', wrap_locale_context(babel_dates.format_date)),
        ('datetime', wrap_locale_context(babel_dates.format_datetime)),
        ('decimal', wrap_locale_context(babel_numbers.format_decimal)),
        ('deeptrans', deeptrans),
        ('expand_partial', expand_partial),
        ('hash', hash_value),
        ('jsonify', jsonify),
        ('markdown', markdown_filter),
        ('number', wrap_locale_context(babel_numbers.format_number)),
        ('percent', wrap_locale_context(babel_numbers.format_percent)),
        ('relative', relative_filter),
        ('re_replace', regex_replace()),
        ('render', render_filter),
        ('shuffle', shuffle_filter),
        ('slug', slug_filter(pod=pod)),
        ('time', wrap_locale_context(babel_dates.format_time)),
    )



"""Features control."""

from grow.common import base_config

class Features(object):
    """Control features."""

    def __call__(self, feature):
        """Ability to call the instance to shortcut to test enabled features."""
        return self.is_enabled(feature)

    def __init__(self, enabled=None, disabled=None, default_enabled=True):
        self._enabled = set()
        self._disabled = set()
        self._config = {}
        self.default_enabled = default_enabled

        if enabled is not None:
            for feature in enabled:
                self.enable(feature)

        if disabled is not None:
            for feature in disabled:
                self.disable(feature)

    def config(self, feature):
        """Configuration for a feature."""
        return self._config.get(feature, base_config.BaseConfig())

    def disable(self, feature):
        """Disable the feature."""
        self._disabled.add(feature)

    def enable(self, feature, config=None):
        """Enable the feature."""
        self._enabled.add(feature)
        self._disabled.discard(feature)
        if config:
            self._config[feature] = base_config.BaseConfig(config=config)

    def is_disabled(self, feature):
        """Determine if the feature is disabled."""
        return not self.is_enabled(feature)

    def is_enabled(self, feature):
        """Determine if the feature is enabled."""
        if feature in self._disabled:
            return False
        if feature in self._enabled:
            return True
        return self.default_enabled is True



"""Subcommand for filtering untranslated messages."""

import os
import click
from grow.commands import shared
from grow.common import rc_config
from grow.pods import pods
from grow import storage


CFG = rc_config.RC_CONFIG.prefixed('grow.translations.filter')


@click.command(name='filter')
@shared.pod_path_argument
@click.option('-o', type=str, default=None,
              help='Where to write the extracted translation catalog. The path'
                   ' must be relative to the pod\'s root.')
@click.option('-f', default=CFG.get('force', False), is_flag=True,
              help='Whether to force an update when writing localized message'
                   ' catalogs.')
@shared.include_header_option(CFG)
@shared.include_obsolete_option(CFG)
@shared.locale_option(
    help_text='Which locale(s) to analyze when creating template catalogs'
              ' that contain only untranslated messages. This option is'
              ' only applicable when using --untranslated.')
@shared.localized_option(CFG)
@shared.out_dir_option(
    CFG, help_text=('Where to write extracted localized translation catalogs.'
                    ' The path must be relative to the pod\'s root. This option'
                    ' is only applicable when using --localized.'))
@shared.path_option
def translations_filter(pod_path, locale, o, include_obsolete, localized, path,
                        include_header, out_dir, f):
    """Filters untranslated messages from catalogs into new catalogs."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    pod = pods.Pod(root, storage=storage.FileStorage)
    with pod.profile.timer('grow_translations_filter'):
        catalogs = pod.get_catalogs()
        if not locale:
            locale = catalogs.list_locales()
        if out_dir and pod.file_exists(out_dir) and not f:
            raise click.UsageError(
                '{} exists. You must specify a directory that does not exist, or '
                'use the "-f" flag, which will force update catalogs within the '
                'specified directory.'.format(out_dir))
        catalogs.filter(out_path=o, out_dir=out_dir,
                        include_obsolete=include_obsolete,
                        localized=localized, paths=path,
                        include_header=include_header, locales=locale)
    return pod



"""Preprocessor for running webpack tasks."""

import os
import atexit
import subprocess
from protorpc import messages
from grow.common import subprocesses
from grow.preprocessors import base
from grow.sdk import sdk_utils


# Keep track of any child processes to terminate on exit.
# pylint: disable=invalid-name
_child_processes = []


class Config(messages.Message):
    """Config for Gulp preprocessor."""
    build_task = messages.StringField(1, default='')
    run_task = messages.StringField(2, default='--watch')
    command = messages.StringField(3, default='webpack')


class WebpackPreprocessor(base.BasePreprocessor):
    """Preprocessor for Gulp."""

    KIND = 'webpack'
    Config = Config

    def _get_command(self, task):
        """Construct the command to run the given webpack task."""
        commands = [self.config.command, task]
        if self.pod.file_exists('/.nvmrc'):
            # Need to source NVM first to get the nvm command to work.
            commands = ['. $NVM_DIR/nvm.sh && nvm exec'] + commands
        return ' '.join(commands)

    def run(self, build=True):
        # Avoid restarting the Gulp subprocess if the preprocessor is
        # being run as a result of restarting the server.
        if 'RESTARTED' in os.environ:
            return
        task = self.config.build_task if build else self.config.run_task
        command = self._get_command(task)
        args = sdk_utils.subprocess_args(self.pod, shell=True)
        process = subprocess.Popen(command, **args)
        _child_processes.append(process)
        if not build:
            return
        code = process.wait()
        if code != 0:
            text = 'Failed to run: {}'.format(command)
            raise base.PreprocessorError(text)


@atexit.register
def _kill_child_process():
    """Sometimes the child process keeps going after grow is done running."""
    subprocesses.kill_child_processes(_child_processes)



"""Pod core extension."""

from grow import extensions
from grow.extensions import hooks


class PodDevFileChangeHook(hooks.DevFileChangeHook):
    """Handle the dev file change hook."""

    # pylint: disable=arguments-differ
    def trigger(self, previous_result, pod_path, *_args, **_kwargs):
        """Trigger the file change hook."""
        if pod_path == '/{}'.format(self.pod.FILE_PODSPEC):
            self.pod.reset_yaml()
        if pod_path == '/{}'.format(self.pod.FILE_EXTENSIONS):
            self.pod.logger.info(
                '{} has changed. Run `grow install` to install updates.'.format(
                    self.pod.FILE_EXTENSIONS))

        if previous_result:
            return previous_result
        return None


# pylint: disable=abstract-method
class PodExtension(extensions.BaseExtension):
    """Extension for handling core pod functionality."""

    @property
    def available_hooks(self):
        """Returns the available hook classes."""
        return [PodDevFileChangeHook]



# NOTE: This is needed so extensions can import `grow` packages.
import logging
import os
import sys
sys.path.extend(
    [os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', '..')])

from grow import commands
from grow.commands import group
commands.add_subcommands(group.grow)

logging.basicConfig(level=logging.INFO, format='%(message)s')


def main(as_module=False):
    args = sys.argv[1:]
    name = None
    if as_module:
        name = 'python -m grow'
        sys.argv = ['-m', 'grow'] + args
    group.grow.main(args=args, prog_name=name)


if __name__ == '__main__':
    main(as_module=True)


"""Base command for grow."""

from grow.deployments.destinations import local as local_destination
import click
import os
import pkg_resources

version = pkg_resources.get_distribution('grow').version

HELP_TEXT = ('Grow is a declarative file-based website generator. Read docs at '
             'https://grow.dev. This is version {}.'.format(version))

# pylint: disable=unused-argument
@click.group(help=HELP_TEXT)
@click.version_option(prog_name='grow', version=version)
@click.option('--auth', help='Information used to sign in to services that'
              ' require authentication. --auth should be an email address.',
              envvar='GROW_AUTH')
@click.option('--clear-auth', default=False, is_flag=True,
              help='Clears stored auth information.')
@click.option('--auth-key-file', help='Path to a private key file used for'
              ' services that require authentication.', envvar='GROW_KEY_FILE')
@click.option(
    '--interactive-auth', default=False, is_flag=True,
    envvar='INTERACTIVE_AUTH',
    help='Whether to automatically open an authorization page in your'
         ' default web browser for any steps that require authentication.'
         ' If you are running Grow on a machine with access to a web browser,'
         ' you may use --interactive-auth to automatically open the web'
         ' browser. By default, this option is turned off, requiring you to'
         ' manually copy and paste an authorization code.')
@click.option('--profile',
              default=False, is_flag=True,
              help='Show report of pod operation timing for performance analysis.')
def grow(auth, clear_auth, auth_key_file, interactive_auth, profile):
    """Grow CLI command."""
    if interactive_auth not in (None, False):
        os.environ['INTERACTIVE_AUTH'] = str(interactive_auth)
    if auth is not None:
        os.environ['AUTH_EMAIL_ADDRESS'] = str(auth)
    if auth_key_file is not None:
        os.environ['AUTH_KEY_FILE'] = str(auth_key_file)
    if clear_auth:
        os.environ['CLEAR_AUTH'] = '1'


@grow.resultcallback()
def process_subcommands(pod, profile, **_):
    """Handle flags that need to process after the sub command."""

    if not pod:
        return

    if profile:
        destination = local_destination.LocalDestination(
            local_destination.Config())
        destination.pod = pod
        destination.export_profile_report()



"""Extension for parsing markdown documents."""

import json
import re
from markdown import extensions
from markdown import preprocessors
from markdown.extensions import toc
from protorpc import messages
from protorpc import protojson
from pygments import highlight
from pygments import lexers
from pygments.formatters import html
from . import utils


def config_from_json(config_class, config):
    config = json.dumps(config)
    return protojson.decode_message(config_class, config)


def get_config(kind, config_class, pod):
    config = config_class()
    if 'markdown' in pod.podspec:
        markdown = pod.podspec.markdown
        if 'extensions' in markdown:
            for extension in markdown['extensions']:
                if extension.get('kind', '') != kind:
                    continue
                return config_from_json(config_class, extension)
    return config


class TocExtension(toc.TocExtension):
    KIND = 'toc'

    class Config(messages.Message):
        marker = messages.StringField(1)
        title = messages.StringField(2)
        baselevel = messages.IntegerField(3)
        anchorlink = messages.BooleanField(4)
        permalink = messages.BooleanField(5)
        separator = messages.StringField(6)
        toc_depth = messages.IntegerField(7)

    def __init__(self, pod):
        config = get_config(TocExtension.KIND, TocExtension.Config, pod)
        config_kwargs = {}
        for item in config.all_fields():
            val = config.get_assigned_value(item.name)
            if val is not None:
                config_kwargs[item.name] = val
        # HTML5 allows all non-space characters for a valid id.
        config_kwargs['slugify'] = lambda value, separator: separator.join(value.split()).lower()
        super(TocExtension, self).__init__(**config_kwargs)


class IncludePreprocessor(preprocessors.Preprocessor):

    REGEX = re.compile(r"^\[include\('([^')]*)'\)\]")

    def __init__(self, pod):
        super(IncludePreprocessor, self).__init__()
        self.pod = pod

    def run(self, lines):
        new_lines = []
        for line in lines:
            pod_paths = IncludePreprocessor.REGEX.findall(line)
            if not pod_paths or line.startswith('    '):
                new_lines.append(line)
            for pod_path in pod_paths:
                doc = self.pod.get_doc(pod_path)
                included_lines = doc.body.split('\n')
                new_lines.extend(included_lines)
        return new_lines


class IncludeExtension(extensions.Extension):

    def __init__(self, pod):
        self.pod = pod

    def extendMarkdown(self, md, md_globals):
        md.registerExtension(self)
        self.processor = IncludePreprocessor(self.pod)
        # Adds the preprocessor to the beginning of the list of preprocessors.
        # https://github.com/waylan/Python-Markdown/blob/master/markdown/odict.py#L7
        md.preprocessors.add('include', self.processor, '_begin')


class UrlPreprocessor(preprocessors.Preprocessor):

    REGEX = re.compile("\[url\('([^']*)'\)\]")

    def __init__(self, pod):
        super(UrlPreprocessor, self).__init__()
        self.pod = pod

    def run(self, lines):
        new_lines = []
        for line in lines:
            pod_paths = UrlPreprocessor.REGEX.findall(line)
            if not pod_paths or line.startswith('    '):
                new_lines.append(line)
            else:
                for pod_path in pod_paths:
                    # Can not import `grow` from within extensions?
                    if pod_path.startswith('/content'):
                        doc = self.pod.get_doc(pod_path)
                    else:
                        doc = self.pod.get_static(pod_path)
                    line = re.sub(
                        UrlPreprocessor.REGEX, doc.url.path, line, count=1)
                new_lines.append(line)
        return new_lines


class UrlExtension(extensions.Extension):

    def __init__(self, pod):
        self.pod = pod

    def extendMarkdown(self, md, md_globals):
        md.registerExtension(self)
        self.processor = UrlPreprocessor(self.pod)
        md.preprocessors.add('url', self.processor, '_begin')


class CodeBlockPreprocessor(preprocessors.Preprocessor):
    """
    Adapted from:
    https://bitbucket.org/birkenfeld/pygments-main/ \
        src/e79a7126551c39d5f8c1b83a79c14e86992155a4/external/markdown-processor.py
    """
    KIND = 'sourcecode'
    pattern_tag = re.compile(
        r'\[sourcecode(:(?P<lang>[^, \]]*))?(, hl_lines=(?P<q>[\'"])(?P<lines>[^\'"]*)(?P=q))?\](?P<content>.+?)\[/sourcecode\]',
        re.S)

    class Config(messages.Message):
        classes = messages.BooleanField(1, default=False)
        class_name = messages.StringField(2, default='code')
        highlighter = messages.StringField(3, default='pygments')
        theme = messages.StringField(4, default='default')

    def __init__(self, pod):
        super(CodeBlockPreprocessor, self).__init__()
        self.pod = pod

    @property
    @utils.memoize
    def formatter(self):
        return self.get_formatter()

    @property
    @utils.memoize
    def config(self):
        return get_config(
            CodeBlockPreprocessor.KIND,
            CodeBlockPreprocessor.Config, self.pod)

    def get_formatter(self, hl_lines=''):
        return html.HtmlFormatter(
            noclasses=(not self.config.classes), cssclass=self.config.class_name,
            style=self.config.theme, hl_lines=hl_lines)

    def run(self, lines):
        class_name = self.config.class_name

        def repl(m):
            language = m.group('lang')
            if language in ['', 'none']:
                language = 'text'
            hl_lines = m.group('lines')
            content = m.group('content')
            if self.config.highlighter == 'pygments':
                formatter = self.formatter
                if hl_lines:
                    formatter = self.get_formatter(hl_lines=hl_lines)
                try:
                    lexer = lexers.get_lexer_by_name(language)
                except ValueError:
                    # pylint: disable=no-member
                    lexer = lexers.TextLexer()
                return '\n{}\n'.format(highlight(content, lexer, formatter))
            elif self.config.highlighter == 'plain':
                text = '\n\n<div class="{}"><pre><code class="{}">{}</code></pre></div>\n\n'
                return text.format(self.config.class_name, language, content)
            text = '{} is an invalid highlighter. Valid choices are: pygments, plain.'
            raise ValueError(text.format(self.config.highlighter))
        content = '\n'.join(lines)
        content = self.pattern_tag.sub(repl, content)
        return content.split('\n')


class CodeBlockExtension(extensions.Extension):

    def __init__(self, pod):
        self.pod = pod

    def extendMarkdown(self, md, md_globals):
        md.registerExtension(self)
        self.processor = CodeBlockPreprocessor(self.pod)
        md.preprocessors.add('sourcecode', self.processor, '_begin')



"""URL utility class."""

import os


class Url(object):
    """Url utility class."""

    def __init__(self, path, host=None, port=None, scheme=None):
        self.path = path
        self.host = 'localhost' if host is None else host
        self.scheme = 'http' if scheme is None else scheme
        default_port = 80
        if self.scheme == 'https':
            default_port = 443
        self.port = default_port if port is None else port
        if scheme is None and self.port == 443:
            self.scheme = 'https'

    def __str__(self):
        url = '{}://{}'.format(self.scheme, self.host)
        if ((self.scheme == 'http' and self.port != 80)
                or (self.scheme == 'https' and self.port != 443)):
            url += ':{}'.format(self.port)
        url += self.path
        return url

    def __eq__(self, other):
        return (
            isinstance(other, Url)
            and self.path == other.path
            and self.host == other.host
            and self.port == other.port
            and self.scheme == other.scheme)

    def __lt__(self, other):
        return str(self) < str(other)

    def __repr__(self):
        return '<Url: {}>'.format(str(self))

    @staticmethod
    def create_relative_path(path, relative_to):
        """Create a relative path to another url."""
        if isinstance(path, Url):
            path = path.path
        if path.startswith(('http://', 'https://')):
            return path

        # Need to support relative paths that are not directories.
        if not path.endswith('/') or not relative_to.endswith('/'):
            if path.endswith('/'):
                path_head = path
                path_tail = None
            else:
                path_head, path_tail = os.path.split(path)

            if relative_to.endswith('/'):
                relative_head = relative_to
            else:
                relative_head, _ = os.path.split(relative_to)

            result = os.path.relpath(path_head, relative_head)
            if path_tail:
                result = '{}/{}'.format(result, path_tail)
        else:
            result = os.path.relpath(path, relative_to)
        if path.endswith('/'):
            result = result + '/'
        if not result.startswith(('/', '.')):
            return './' + result
        return result



"""Grow deployment management."""

from protorpc import protojson
import json


class Deployments:
    """Manage the deployment destinations."""

    def __init__(self):
        self._deployments = {}

    @staticmethod
    def config_from_json(deployment_cls, content):
        """Create a deployment config from json content."""
        config_cls = deployment_cls.Config
        return protojson.decode_message(config_cls, content)

    def register_destination(self, destination_cls):
        """Register or override a deployment destination."""
        self._deployments[destination_cls.KIND] = destination_cls

    def make_deployment(self, kind, config, name='default'):
        """Create an instance of the deployment destination."""
        destination_cls = self._deployments.get(kind)
        if destination_cls is None:
            raise ValueError('No deployment destination exists for "{}".'.format(kind))

        if isinstance(config, dict):
            config = json.dumps(config)
        if not config:
            config = '{}'
        config = self.config_from_json(destination_cls, config)
        return destination_cls(config, name=name)



"""An Env holds the environment context that a pod is running in."""

import time
from protorpc import messages
from grow.common import urls


class Name(object):
    DEV = 'dev'


class EnvConfig(messages.Message):
    host = messages.StringField(1)
    scheme = messages.StringField(2)
    port = messages.IntegerField(3)
    name = messages.StringField(4)
    cached = messages.BooleanField(5, default=True)
    fingerprint = messages.StringField(6)
    dev = messages.BooleanField(7, default=False)


class Env(object):

    def __init__(self, config):
        self.name = config.name
        self.config = config
        self.cached = config.cached
        self.fingerprint = config.fingerprint or str(int(time.time()))

    def __repr__(self):
        return '<Env: {}>'.format(self.url)

    @property
    def host(self):
        return self.config.host or 'localhost'

    @host.setter
    def host(self, value):
        self.config.host = value

    @property
    def port(self):
        return self.config.port

    @port.setter
    def port(self, value):
        self.config.port = value

    @property
    def scheme(self):
        return self.config.scheme

    @scheme.setter
    def scheme(self, value):
        self.config.scheme = value

    @property
    def dev(self):
        return self.config.dev

    @property
    def url(self):
        return urls.Url(
            path='/',
            host=self.host,
            port=self.port,
            scheme=self.scheme)

    def to_wsgi_env(self):
        return {
            'REQUEST_METHOD': 'GET',
            'SERVER_NAME': self.host,
            'SERVER_PORT': str(self.port),
            'wsgi.url_scheme': self.scheme,
        }



"""Gerrit installer class."""

import subprocess
import urllib.parse
from grow.common import utils
from grow.sdk.installers import base_installer

KNOWN_GERRIT_HOSTS = (
    'googlesource.com',
)


class GerritInstaller(base_installer.BaseInstaller):
    """Gerrit installer."""

    KIND = 'gerrit'

    @property
    def should_run(self):
        """Should the installer run?"""
        gerrit_setting = self.config.get('gerrit', None)
        if gerrit_setting is not None:
            return gerrit_setting

        repo = utils.get_git_repo(self.pod.root)
        if repo is None:
            return False
        for remote in repo.remotes:
            url = remote.config_reader.get('url')
            result = urllib.parse.urlparse(url)
            if result.netloc.endswith(KNOWN_GERRIT_HOSTS):
                return True
        return False

    def install(self):
        """Install dependencies."""
        curl_command = (
            'curl -sLo '
            '`git rev-parse --git-dir`/hooks/commit-msg '
            'https://gerrit-review.googlesource.com/tools/hooks/commit-msg')
        process = subprocess.Popen(
            curl_command, **self.subprocess_args(shell=True))
        code = process.wait()
        if not code:
            chmod_command = 'chmod +x `git rev-parse --git-dir`/hooks/commit-msg'
            process = subprocess.Popen(
                chmod_command, **self.subprocess_args(shell=True))
            code = process.wait()
            if not code:
                return
            raise base_installer.InstallError(
                'There was an error running `{}`.'.format(chmod_command))
        raise base_installer.InstallError(
            'There was an error running `{}`.'.format(curl_command))



"""Routes core extension."""

import os
from werkzeug import wrappers
from grow import extensions
from grow.collections import collection
from grow.common import timer
from grow.extensions import hooks
from grow.performance import docs_loader
from grow.pods import ui
from grow.routing import router as grow_router
from grow.server import handlers


class RoutesDevHandlerHook(hooks.DevHandlerHook):
    """Handle the dev handler hook."""

    @staticmethod
    def _create_response(pod, routes, title, concrete=True):
        env = ui.create_jinja_env()
        template = env.get_template('views/base.html')
        kwargs = {
            'pod': pod,
            'partials': [{
                'partial': 'routes',
                'concrete': concrete,
                'routes': routes,
            }],
            'title': title,
        }
        content = template.render(kwargs)
        response = wrappers.Response(content)
        response.headers['Content-Type'] = 'text/html'
        return response

    @staticmethod
    def serve_routes_abstract(pod, _request, _matched, **_kwargs):
        """Handle the request for abstract routes."""
        routes = pod.router.routes
        return RoutesDevHandlerHook._create_response(
            pod, routes, 'Pod Abstract Routes')

    @staticmethod
    def serve_routes_concrete(pod, _request, _matched, **_kwargs):
        """Handle the request for routes."""
        router = grow_router.Router(pod)
        router.use_simple()
        router.add_all(concrete=True, use_cache=False)
        routes = router.routes
        return RoutesDevHandlerHook._create_response(
            pod, routes, 'Pod Routes')

    # pylint: disable=arguments-differ
    def trigger(self, previous_result, routes, *_args, **_kwargs):
        """Execute dev handler modification."""
        routes.add('/_grow/ui/tools/:tool', grow_router.RouteInfo(
            'console', meta={
                'handler': handlers.serve_ui_tool,
            }))
        routes.add('/_grow', grow_router.RouteInfo('console', meta={
            'handler': handlers.serve_console,
        }))

        routes.add('/_grow/routes', grow_router.RouteInfo(
            'console', meta={
                'handler': RoutesDevHandlerHook.serve_routes_concrete,
            }))
        routes.add('/_grow/routes/concrete', grow_router.RouteInfo(
            'console', meta={
                'handler': RoutesDevHandlerHook.serve_routes_concrete,
            }))
        routes.add('/_grow/routes/abstract', grow_router.RouteInfo(
            'console', meta={
                'handler': RoutesDevHandlerHook.serve_routes_abstract,
            }))


class RoutesDevFileChangeHook(hooks.DevFileChangeHook):
    """Handle the dev file change hook."""

    @staticmethod
    def _reset_routes(pod):
        with timer.Timer() as router_time:
            pod.router.routes.reset()
            pod.router.add_all(concrete=False, use_cache=False)
            # Trigger the dev handler hook.
            pod.extensions_controller.trigger(
                'dev_handler', pod.router.routes)
        pod.logger.info('{} routes rebuilt in {:.3f} s'.format(
            len(pod.router.routes), router_time.secs))

    def trigger(self, previous_result, pod_path, *_args, **_kwargs):
        """Trigger the file change hook."""
        pod = self.pod
        basename = os.path.basename(pod_path)
        ignore_doc = basename.startswith(collection.Collection.IGNORE_INITIAL)

        if pod_path == '/{}'.format(pod.FILE_PODSPEC):
            self._reset_routes(pod)
        elif (pod_path.endswith(collection.Collection.BLUEPRINT_PATH)
              and pod_path.startswith(collection.Collection.CONTENT_PATH)):
            self._reset_routes(pod)
        elif pod_path.startswith(collection.Collection.CONTENT_PATH) and not ignore_doc:
            trigger_doc = pod.get_doc(pod_path)
            col = trigger_doc.collection
            base_docs = []
            original_docs = []
            trigger_docs = col.list_servable_document_locales(pod_path)

            for dep_path in pod.podcache.dependency_graph.get_dependents(pod_path):
                base_docs.append(pod.get_doc(dep_path))
                original_docs += col.list_servable_document_locales(dep_path)

            # Normally this would be part of the podcache extension.
            # Needs to be between retrieving the original docs and flushing
            # the cache to be able to compare paths.
            for doc in base_docs:
                pod.podcache.document_cache.remove(doc)
                pod.podcache.collection_cache.remove_document_locales(doc)

            # Force load the docs and fix locales.
            docs_loader.DocsLoader.load(pod, base_docs, ignore_errors=True)
            docs_loader.DocsLoader.fix_default_locale(
                pod, base_docs, ignore_errors=True)

            # The routing map should remain unchanged most of the time.
            added_docs = []
            removed_docs = []
            for original_doc in original_docs:
                # Removed documents should be removed.
                if not original_doc.exists:
                    removed_docs.append(original_doc)
                    continue

                updated_doc = pod.get_doc(
                    original_doc.pod_path, original_doc._locale_kwarg)

                # When the serving path has changed, updated in routes.
                if (updated_doc.has_serving_path()
                        and original_doc.get_serving_path() != updated_doc.get_serving_path()):
                    added_docs.append(updated_doc)
                    removed_docs.append(original_doc)

                # If the locales change then we need to adjust the routes.
                original_locales = set([str(l) for l in original_doc.locales])
                updated_locales = set([str(l) for l in updated_doc.locales])

                new_locales = updated_locales - original_locales
                for locale in new_locales:
                    new_doc = pod.get_doc(original_doc.pod_path, locale)
                    if new_doc.has_serving_path() and new_doc not in added_docs:
                        added_docs.append(new_doc)

                removed_locales = original_locales - updated_locales
                for locale in removed_locales:
                    removed_doc = pod.get_doc(original_doc.pod_path, locale)
                    if removed_doc.has_serving_path():
                        if removed_doc not in removed_docs:
                            removed_docs.append(removed_doc)

            # Check for new docs.
            for trigger_doc in trigger_docs:
                if trigger_doc.exists and trigger_doc.has_serving_path():
                    if not pod.router.routes.match(trigger_doc.get_serving_path()):
                        added_docs.append(trigger_doc)
            if added_docs or removed_docs:
                pod.router.reconcile_documents(
                    remove_docs=removed_docs, add_docs=added_docs)

        if previous_result:
            return previous_result
        return None


# pylint: disable=abstract-method
class RoutesExtension(extensions.BaseExtension):
    """Extension for handling core routes functionality."""

    @property
    def available_hooks(self):
        """Returns the available hook classes."""
        return [RoutesDevHandlerHook, RoutesDevFileChangeHook]



"""Shared options or arguments for cli commands."""

import click
from grow.common import rc_config


CFG = rc_config.RC_CONFIG.prefixed('grow.shared')


def pod_path_argument(func):
    """Argument for pod path."""
    return click.argument('pod_path', default='.')(func)


def deployment_option(config):
    """Option for changing env based on deployment name."""
    shared_default = CFG.get('deployment', None)
    config_default = config.get('deployment', shared_default)

    def _decorator(func):
        return click.option(
            '--deployment', default=config_default,
            help='Name of the deployment config to use.')(func)
    return _decorator


def force_untranslated_option(config):
    """Option for forcing untranslated string deployment."""
    shared_default = CFG.get('force-untranslated', False)
    config_default = config.get('force-untranslated', shared_default)

    def _decorator(func):
        return click.option(
            '--force-untranslated', 'force_untranslated',
            default=config_default, is_flag=True,
            help='Whether to force untranslated strings to be uploaded.')(func)
    return _decorator


def include_header_option(config):
    """Option for preserving catalog header."""
    shared_default = CFG.get('include-header', False)
    config_default = config.get('include-header', shared_default)

    def _decorator(func):
        return click.option(
            '--include-header',
            default=config_default, is_flag=True,
            help='Whether to preserve headers at the beginning of catalogs.')(func)
    return _decorator


def include_obsolete_option(config, default_value=None):
    """Option for including obsolete translations."""
    shared_default = CFG.get('include-obsolete', default_value)
    config_default = config.get('include-obsolete', shared_default)

    help_text = ('Whether to include obsolete messages. If false, obsolete'
                 ' messages will be removed from the catalog template.')

    if default_value is False:
        help_text = '{} {}'.format(
            help_text, 'By default, Grow cleans obsolete messages from the catalog template.')

    def _decorator(func):
        return click.option(
            '--include-obsolete/--no-include-obsolete',
            default=config_default, is_flag=True, help=help_text)(func)
    return _decorator


def locale_option(help_text=None, multiple=True):
    """Option for providing locale(s)."""
    if help_text is None:
        help_text = 'Locale to target.'

    def _decorator(func):
        return click.option(
            '--locale', type=str, multiple=multiple, help=help_text)(func)
    return _decorator


def localized_option(config):
    """Option for localizing catalogs."""
    shared_default = CFG.get('localized', None)
    config_default = config.get('localized', shared_default)

    def _decorator(func):
        return click.option(
            '--localized/--no-localized',
            is_flag=True, default=config_default,
            help='Whether to create localized message catalogs. Use this'
                 ' option if content varies by locale.')(func)
    return _decorator


def out_dir_option(config, help_text=None):
    """Option for output directory."""
    shared_default = CFG.get('out-dir', None)
    config_default = config.get('out-dir', shared_default)

    if help_text is None:
        help_text = 'Directory to output files.'

    def _decorator(func):
        return click.option(
            '--out-dir', '--out_dir',
            type=str, default=config_default, help=help_text)(func)
    return _decorator


def path_option(func):
    """Option for specifying a path for extraction."""
    return click.option(
        '--path',
        type=str, multiple=True,
        help='Which paths to extract strings from. By default, all paths'
             ' are extracted. This option is useful if you\'d like to'
             ' generate a partial messages file representing just a'
             ' specific set of files.')(func)


def preprocess_option(config):
    """Option for running preprocessors."""
    shared_default = CFG.get('preprocess', True)
    config_default = config.get('preprocess', shared_default)

    def _decorator(func):
        return click.option(
            '--preprocess/--no-preprocess', '-p/-np',
            is_flag=True, default=config_default,
            help='Whether to run preprocessors.')(func)
    return _decorator


def routes_file_option(help_text=None):
    """Option for providing a routes file instead of pulling from content."""
    if help_text is None:
        help_text = 'Use routes file to load routes instead of loading from files.'

    def _decorator(func):
        return click.option(
            '--routes-file', '--routes_file', type=str, default=None,
            help=help_text)(func)
    return _decorator


def service_option(func):
    """Option for configuring the transltor service to use."""
    return click.option('--service', '-s', type=str,
                        help='Name of the translator service to use. This option is'
                        ' only required if more than one service is configured.')(func)


def shard_option(func):
    """Option for configuring the current shard."""
    return click.option(
        '--shard', type=int,
        help='Current index of shard being used for sharding.')(func)


def shards_option(func):
    """Option for configuring the shard count."""
    return click.option(
        '--shards', type=int,
        help='Number of shards being used for sharding.')(func)


def threaded_option(config):
    """Option for using threading when rendering."""
    shared_default = CFG.get('threaded', True)
    config_default = config.get('threaded', shared_default)

    def _decorator(func):
        return click.option(
            '--threaded/--no-threaded', '-t/-nt', is_flag=True,
            default=config_default,
            help='Use threading during rendering pipeline.')(func)
    return _decorator


def work_dir_option(func):
    """Option for working directory."""
    help_text = 'Directory to pull working files from.'
    return click.option(
        '--work-dir', '--work_dir', type=str, default=None,
        help=help_text)(func)



"""Custom importing for pod extensions."""

import importlib.util
import os
import sys
import threading


class ExtensionImporter:
    """Custom find the extensions based on the pod path."""

    @staticmethod
    def find_extension(extension_ref, pod_root):
        """Using the pod path as the base to find the extension module."""
        extension_parts = extension_ref.split('.')
        extension_class_name = extension_parts.pop()
        try:
            module_name = extension_parts[-1]
        except IndexError:
            raise ImportError('Unable to find extension: {}'.format(extension_ref))

        # Try first as the full module path.
        module_path = '{}{sep}{}.py'.format(
            pod_root, '/'.join(extension_parts), sep=os.path.sep)

        if not os.path.exists(module_path):
            module_path = '{}{sep}{}{sep}__init__.py'.format(
                pod_root, '/'.join(extension_parts), sep=os.path.sep)

        if not os.path.exists(module_path):
            module_path = '{}{sep}{}{sep}{}.py'.format(
                pod_root, '/'.join(extension_parts), module_name, sep=os.path.sep)

        if not os.path.exists(module_path):
            raise ImportError(
                'Unable to find extension module for {!r}'.format(extension_ref))

        spec = importlib.util.spec_from_file_location(module_name, module_path)

        if not spec:
            raise ImportError(
                'Unable to load extension from {!r}'.format(extension_ref))

        # Lock the module import to prevent concurrency issues between pods.
        with threading.Lock():
            # Import the module from the spec and execute to get access.
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            try:
                del sys.modules[module_name]
            except KeyError:
                pass

            # Make a list of the keys since the modules may change when
            # using the keys iterator.
            module_keys = list(sys.modules.keys())
            ext_prefix = '{}.'.format(module_name)
            for key in filter(lambda x: x.startswith(ext_prefix), module_keys):
                try:
                    del sys.modules[key]
                except KeyError:
                    pass

        return getattr(module, extension_class_name)



"""Custom jinja loaders."""

import os
import jinja2


class FrontMatterLoader(jinja2.FileSystemLoader):
    """Load templates that have possible front matter and strip it off."""

    SEPARATOR = '---'

    def get_source(self, environment, template):
        source, filename, uptodate = super(FrontMatterLoader, self).get_source(
            environment, template)
        # For jinja rendering strip off the front matter.
        if self.SEPARATOR in source:
            source_lines = source.splitlines(True)
            separator_index = None
            for index, line in enumerate(source_lines):
                # Look for the separator, ignoring the first line.
                if line.strip() == self.SEPARATOR and index > 0:
                    separator_index = index
                    break
            if separator_index:
                fill_lines = [os.linesep] * (separator_index + 1)
                source = u''.join(fill_lines + source_lines[separator_index+1:])
        return source, filename, uptodate



"""Podcache core extension."""

from grow import extensions
from grow.cache import podcache
from grow.collections import collection
from grow.extensions import hooks


class PodcacheDevFileChangeHook(hooks.DevFileChangeHook):
    """Handle the dev file change hook."""

    # pylint: disable=arguments-differ
    def trigger(self, previous_result, pod_path, *_args, **kwargs):
        """Trigger the file change hook."""

        # We don't want to write cache files when just serving.
        write_cache_file = True
        if 'write_cache_file' in kwargs:
            write_cache_file = kwargs['write_cache_file']

        # Remove any raw file in the cache.
        self.pod.podcache.file_cache.remove(pod_path)

        if pod_path == '/{}'.format(self.pod.FILE_PODSPEC):
            self.pod.podcache.reset()
        elif (pod_path.endswith(collection.Collection.BLUEPRINT_PATH)
              and pod_path.startswith(collection.Collection.CONTENT_PATH)):
            doc = self.pod.get_doc(pod_path)
            self.pod.podcache.collection_cache.remove_collection(doc.collection)
        elif pod_path == '/{}'.format(podcache.FILE_OBJECT_CACHE):
            self.pod.podcache.update(obj_cache=self.pod._parse_object_cache_file())
            if write_cache_file and self.pod.podcache.is_dirty:
                self.pod.logger.info('Object cache changed, updating with new data.')
                self.pod.podcache.write()
        else:
            # Clear any documents that depend on the pod_path.
            for dep_path in self.pod.podcache.dependency_graph.get_dependents(pod_path):
                doc = self.pod.get_doc(dep_path)
                self.pod.podcache.document_cache.remove(doc)
                self.pod.podcache.collection_cache.remove_document_locales(doc)

        if previous_result:
            return previous_result
        return None


# pylint: disable=abstract-method
class PodcacheExtension(extensions.BaseExtension):
    """Extension for handling core podcache functionality."""

    @property
    def available_hooks(self):
        """Returns the available hook classes."""
        return [PodcacheDevFileChangeHook]



"""Subcommand for displaying all routes of a pod."""

import json
import os
import click
from grow.commands import shared
from grow.common import rc_config
from grow.pods import pods
from grow import storage


CFG = rc_config.RC_CONFIG.prefixed('grow.routes')


@click.command(name='routes')
@shared.deployment_option(CFG)
@shared.routes_file_option('Export route information to a json file.')
@shared.pod_path_argument
def inspect_routes(pod_path, routes_file, deployment):
    """Lists routes handled by a pod."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    pod = pods.Pod(root, storage=storage.FileStorage)
    if deployment:
        deployment_obj = pod.get_deployment(deployment)
        pod.set_env(deployment_obj.config.env)
    with pod.profile.timer('grow_inspect_routes'):
        out = []
        pod.router.use_simple()
        pod.router.add_all()

        if routes_file:
            # print pod.router.routes.export()
            pod.write_file(
                routes_file,
                json.dumps(
                    pod.router.routes.export(), sort_keys=True, indent=2,
                    separators=(',', ': ')))
            print('Exported routes to file: {}'.format(routes_file))
            return pod

        for path, node, _ in pod.router.routes.nodes:
            out.append('{} [{}]\n    {}\n'.format(path, node.kind, node.meta))
        click.echo_via_pager('\n'.join(out))
    return pod



"""Grow.dev setup."""

import re
from setuptools import find_packages
from setuptools import setup


DEP_RE = re.compile(r'([\S]*)\s?=\s? [\"\']?([^\'\"]*)[\"\']?', re.IGNORECASE)
INSTALL_REQ = []

with open('Pipfile') as pipfile:
    in_dep_section = False
    for line in pipfile.readlines():
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        if in_dep_section:
            if line.startswith('['):
                in_dep_section = False
                continue
            line_match = DEP_RE.match(line)
            if line_match:
                INSTALL_REQ.append(
                    '{}{}'.format(line_match.group(1).strip('"'), line_match.group(2)))
        elif line == '[packages]':
            in_dep_section = True


setup(
    name='grow',
    version='2.0.6',
    description=(
        'Develop everywhere and deploy anywhere: a declarative '
        'site generator for rapid, high-quality web site production.'
    ),
    long_description=open('description.txt').read().strip(),
    url='https://grow.dev',
    zip_safe=False,
    license='MIT',
    author='Grow.dev Authors',
    author_email='code@grow.dev',
    include_package_data=True,
    packages=find_packages(exclude=[
        'lib*',
        'node_modules',
    ]),
    install_requires=INSTALL_REQ,
    python_requires='>=3.3',
    entry_points="""
        [console_scripts]
        grow=grow.cli:main
    """,
    keywords=[
        'grow',
        'cms',
        'static site generator',
        'content management'
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ])



"""Non-interactive progressbar fallback for non-interactive shells."""

import os
import sys
import datetime
import progressbar


class NonInteractiveProgressBar(object):
    """Non-interactive stub version of progress bar."""

    def __init__(self, message, max_value=None, poll_interval=5, widgets=None, *_args, **kwargs):
        self._last_print = datetime.datetime.now()
        self._last_update = datetime.datetime.now()
        self._min_delta = datetime.timedelta(0, poll_interval)
        self._max_value = max_value
        self._widgets = widgets or []
        self._data = kwargs
        self._started_on = None
        self._finished_on = None
        self._value = 0

        print(message)

    @property
    def percentage(self):
        """Percentage of value vs max_value."""
        if not self._max_value:
            return None
        return self._value / self._max_value

    @property
    def value(self):
        """Value of the progress bar."""
        return self._value

    def data(self):
        """Try to mimic the data available to the progressbar."""
        elapsed = self._last_update - self._started_on
        return dict(
            # The maximum value (can be None with iterators)
            max_value=self._max_value,
            # Start time of the widget
            start_time=self._started_on,
            # Last update time of the widget
            last_update_time=self._last_update,
            # End time of the widget
            end_time=self._finished_on,
            # The current value
            value=self._value,
            # The seconds since the bar started
            total_seconds_elapsed=elapsed.total_seconds,
            # The seconds since the bar started modulo 60
            seconds_elapsed=(elapsed.seconds % 60) +
            (elapsed.microseconds / 1000000.),
            # The minutes since the bar started modulo 60
            minutes_elapsed=(elapsed.seconds / 60) % 60,
            # The hours since the bar started modulo 24
            hours_elapsed=(elapsed.seconds / (60 * 60)) % 24,
            # The hours since the bar started
            days_elapsed=(elapsed.seconds / (60 * 60 * 24)),
            # The raw elapsed `datetime.timedelta` object
            time_elapsed=elapsed,
            # Percentage as a float or `None` if no max_value is available
            percentage=self.percentage,
        )

    def finish(self):
        """Finished with the progressbar."""
        self._finished_on = datetime.datetime.now()
        for widget in self._widgets:
            print(widget(self, self.data()))

    def start(self):
        """Started with the progressbar."""
        self._started_on = datetime.datetime.now()

    def update(self, value):
        self._last_update = datetime.datetime.now()
        self._value = value
        if self._last_update - self._last_print > self._min_delta:
            for widget in self._widgets:
                print(widget(self, self.data()))
            self._last_print = self._last_update


def create_progressbar(message, *args, **kwargs):
    """Create the correct progressbar based on availability of interactive shell."""
    if 'CI' in os.environ and os.environ['CI']:
        kwargs['poll_interval'] = 10
    return progressbar.ProgressBar(*args, **kwargs)
    # Removing for now to see if it is causing an issue with
    # if hasattr(sys.stdout, 'fileno') and os.isatty(sys.stdout.fileno()):
    #     return progressbar.ProgressBar(*args, **kwargs)
    # return NonInteractiveProgressBar(message, *args, **kwargs)



"""Local deployment destination."""

import errno
import os
import shutil
from protorpc import messages
from grow.deployments.destinations import base
from grow.pods import env
from grow.routing import router
from grow.storage import file_storage


class Config(messages.Message):
    out_dir = messages.StringField(1, default='')
    env = messages.MessageField(env.EnvConfig, 2)
    keep_control_dir = messages.BooleanField(3, default=False)
    before_deploy = messages.StringField(4, repeated=True)
    after_deploy = messages.StringField(5, repeated=True)
    control_dir = messages.StringField(6)
    filters = messages.MessageField(router.FilterConfig, 7, repeated=True)
    tags = messages.StringField(8, repeated=True)


class LocalDestination(base.BaseDestination):
    KIND = 'local'
    Config = Config
    storage = file_storage.FileStorage

    def __str__(self):
        return os.path.abspath(os.path.join(self.out_dir))

    @property
    def out_dir(self):
        return os.path.expanduser(self.config.out_dir)

    def read_file(self, path):
        path = os.path.join(self.out_dir, path.lstrip('/'))
        return self.storage.read(path)

    def delete_file(self, path):
        out_path = os.path.join(self.out_dir, path.lstrip('/'))
        try:
            self.storage.delete(out_path)
        except IOError as e:
            if e.errno == errno.ENOENT:
                pass
            else:
                raise

    def write_file(self, rendered_doc):
        path = rendered_doc.path
        out_path = os.path.join(self.out_dir, path.lstrip('/'))
        if rendered_doc.file_path:
            dir_name = os.path.dirname(out_path)
            try:
                os.makedirs(dir_name)
            except OSError as exc:
                if exc.errno == errno.EEXIST:
                    pass
                else:
                    raise
            shutil.copyfile(rendered_doc.file_path, out_path)
        else:
            self.storage.write(out_path, rendered_doc.read())

    def prelaunch(self, dry_run=False):
        for command in self.config.before_deploy:
            self.command(command)
        super(LocalDestination, self).prelaunch(dry_run)

    def postlaunch(self, dry_run=False):
        if self.success:
            for command in self.config.after_deploy:
                self.command(command)
        super(LocalDestination, self).postlaunch(dry_run)



"""Command for building pods into static deployments."""

import os
import click
from grow.commands import shared
from grow.common import bulk_errors
from grow.common import rc_config
from grow.common import utils
from grow.deployments import stats
from grow.deployments.destinations import local as local_destination
from grow.extensions import hooks
from grow.performance import docs_loader
from grow.pods import pods
from grow.rendering import renderer
from grow import storage


CFG = rc_config.RC_CONFIG.prefixed('grow.build')


# pylint: disable=too-many-locals
@click.command()
@shared.pod_path_argument
@click.option('--clear-cache',
              default=CFG.get('clear-cache', False), is_flag=True,
              help='Clear the pod cache before building.')
@click.option('--file', '--pod-path', 'pod_paths',
              help='Build only pages affected by content files.', multiple=True)
@click.option('--locate-untranslated',
              default=CFG.get('locate-untranslated', False), is_flag=True,
              help='Shows untranslated message information.')
@shared.locale_option(help_text='Filter build routes to specific locale.')
@shared.deployment_option(CFG)
@shared.out_dir_option(CFG)
@shared.preprocess_option(CFG)
@shared.threaded_option(CFG)
@shared.shards_option
@shared.shard_option
@shared.work_dir_option
@shared.routes_file_option()
def build(pod_path, out_dir, preprocess, clear_cache, pod_paths,
          locate_untranslated, deployment, threaded, locale, shards, shard,
          work_dir, routes_file):
    """Generates static files and dumps them to a local destination."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    out_dir = out_dir or os.path.join(root, 'build')

    pod = pods.Pod(root, storage=storage.FileStorage)
    if not pod_paths or clear_cache:
        # Clear the cache when building all, only force if the flag is used.
        pod.podcache.reset(force=clear_cache)
    deployment_obj = None
    if deployment:
        deployment_obj = pod.get_deployment(deployment)
        pod.set_env(deployment_obj.config.env)
    if preprocess:
        with pod.profile.timer('grow_preprocess'):
            pod.preprocess()
    if locate_untranslated:
        pod.enable(pod.FEATURE_TRANSLATION_STATS)
    try:
        with pod.profile.timer('grow_build'):
            config = local_destination.Config(out_dir=out_dir)
            # When using a specific deployment env need to also copy over.
            if deployment_obj:
                config.env = deployment_obj.config.env
            destination = local_destination.LocalDestination(config)
            destination.pod = pod
            repo = utils.get_git_repo(pod.root)
            pod.router.use_simple()
            is_partial = bool(pod_paths) or bool(locale)
            if pod_paths:
                pod_paths = [pod.clean_pod_path(path) for path in pod_paths]
                pod.router.add_pod_paths(pod_paths)
            elif routes_file:
                pod.router.from_data(pod.read_json(routes_file))
            else:
                pod.router.add_all()
            if locale:
                pod.router.filter('whitelist', locales=list(locale))

            # Shard the routes when using sharding.
            if shards and shard:
                is_partial = True
                pod.router.shard(shards, shard)

            if not work_dir:
                # Preload the documents used by the paths after filtering.
                docs_loader.DocsLoader.load_from_routes(pod, pod.router.routes)

            paths = pod.router.routes.paths
            content_generator = renderer.Renderer.rendered_docs(
                pod, pod.router.routes, source_dir=work_dir,
                use_threading=threaded)
            content_generator = hooks.generator_wrapper(
                pod, 'pre_deploy', content_generator, 'build')
            stats_obj = stats.Stats(pod, paths=paths)
            destination.deploy(
                content_generator, stats=stats_obj, repo=repo, confirm=False,
                test=False, is_partial=is_partial)
            pod.podcache.write()
    except bulk_errors.BulkErrors as err:
        # Write the podcache files even when there are rendering errors.
        pod.podcache.write()
        bulk_errors.display_bulk_errors(err)
        raise click.Abort()
    except pods.Error as err:
        raise click.ClickException(str(err))
    if locate_untranslated:
        pod.translation_stats.pretty_print()
        destination.export_untranslated_catalogs()
    return pod



"""Document formatting specifics for parsing and working with documents."""

import collections
import logging
from grow.common import utils
from grow.documents import document_front_matter as doc_front_matter


# Set markdown logging level to info.
logging.getLogger('MARKDOWN').setLevel(logging.INFO)


BOUNDARY_SEPARATOR = '---'


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class BadFormatError(Error, ValueError):
    pass


class BadLocalesError(BadFormatError):
    pass


class DocumentFormat(object):
    """
    Document formatting specifics for parsing and working with documents.

    Defines how to handle documents formatted in various syntax formats.
    """

    def __init__(self, doc):
        self._doc = doc
        self._content = None
        self._raw_content = None

    @staticmethod
    def from_doc(*args, **kwargs):
        doc = kwargs.get('doc', None)
        if not doc:
            raise BadFormatError(
                'Missing `doc` keyword argument for creating format')
        if doc.ext in ('.html', '.htm'):
            return HtmlDocumentFormat(*args, **kwargs)
        if doc.ext in ('.markdown', '.mdown', '.mkdn', '.mkd', '.md'):
            return MarkdownDocumentFormat(*args, **kwargs)
        if doc.ext in ('.yaml', '.yml'):
            return YamlDocumentFormat(*args, **kwargs)
        return TextDocumentFormat(*args, **kwargs)

    @staticmethod
    def format_doc(front_matter, content):
        if front_matter and content:
            return '{0}\n{1}\n{0}\n{2}\n'.format(
                BOUNDARY_SEPARATOR, front_matter.strip(), content.strip())
        elif front_matter:
            return '{}\n'.format(front_matter.strip())
        return '{}\n'.format(content.strip())

    def _parse_content(self):
        """Parse the content from the raw content."""
        _, parsed_content = doc_front_matter.DocumentFrontMatter\
            .split_front_matter(self.raw_content)
        return parsed_content

    def _parse_front_matter(self):
        """Parse the front matter from the raw content."""
        return doc_front_matter.DocumentFrontMatter(
            self._doc)

    @property
    def content(self):
        """Lazy load the content after checking the content cache."""
        if self._content:
            return self._content
        self._content = self._parse_content()
        return self._content

    @utils.cached_property
    def front_matter(self):
        cached_front_matter = self._doc.pod.podcache.document_cache\
            .get_property(self._doc, 'front_matter')
        if cached_front_matter:
            return doc_front_matter.DocumentFrontMatter(
                self._doc, raw_front_matter=cached_front_matter)

        front_matter = self._parse_front_matter()
        self._doc.pod.podcache.document_cache.add_property(
            self._doc, 'front_matter', front_matter.export())
        return front_matter

    @property
    def raw_content(self):
        if self._raw_content:
            return self._raw_content
        if self._doc.exists:
            self._raw_content = self._doc.pod.read_file(self._doc.pod_path)
        return self._raw_content

    @utils.cached_property
    def formatted(self):
        return self.content

    def organize_fields(self, fields):
        """Structure the fields data to keep some minimal structure."""
        new_fields = collections.OrderedDict()

        # Deep sort all fields by default.
        def _walk_field(item, key, node, parent_node):
            try:
                value = node[key]
                new_value = collections.OrderedDict()
                for sub_key in sorted(value.keys()):
                    new_value[sub_key] = value[sub_key]
                node[key] = new_value
            except:
                pass
        utils.walk(fields, _walk_field)

        # Organization rules:
        # $ prefixed fields should come first.
        # Partials key is last.
        # Partials' partial key should be first in partial data.
        # Sort the fields to keep consistent between saves.

        other_keys = []
        for key in sorted(fields.keys()):
            if key.startswith('$'):
                new_fields[key] = fields[key]
            elif key == 'partials':
                pass
            else:
                other_keys.append(key)

        for key in other_keys:
            new_fields[key] = fields[key]

        if 'partials' in fields:
            new_partials = []

            for partial in fields['partials']:
                new_partial = collections.OrderedDict()

                try:
                    # Put the partial name first for easy readability.
                    if 'partial' in partial:
                        new_partial['partial'] = partial['partial']

                    for key in sorted(partial.keys()):
                        if key != 'partial':
                            new_partial[key] = partial[key]

                    new_partials.append(new_partial)
                except TypeError:
                    # When unable to sort the partial keys, use original.
                    new_partials.append(partial)

            new_fields['partials'] = new_partials

        return new_fields

    def to_raw_content(self):
        """Formats the front matter and content into a raw_content string."""
        raw_front_matter = self.front_matter.export()
        return self.format_doc(raw_front_matter, self.content)

    def update(self, fields=utils.SENTINEL, content=utils.SENTINEL):
        """Updates content and frontmatter."""
        if fields is not utils.SENTINEL:
            # Organize some of the fields for minimal consistency.
            fields = self.organize_fields(fields)

            raw_front_matter = utils.dump_yaml(fields)
            self.front_matter.update_raw_front_matter(raw_front_matter)
            self._doc.pod.podcache.document_cache.add_property(
                self._doc, 'front_matter', self.front_matter.export())

        if content is not utils.SENTINEL:
            self._content = content

        self._raw_content = self.to_raw_content()


class HtmlDocumentFormat(DocumentFormat):

    @utils.cached_property
    def formatted(self):
        val = self.content
        return val if val is not None else None


class MarkdownDocumentFormat(DocumentFormat):

    @utils.cached_property
    def markdown(self):
        """Instance of pod flavored markdown."""
        return self._doc.pod.markdown

    @property
    def toc(self):
        """Markdown TOC extension."""
        # Make sure that the document conversion has happened.
        _ = self.formatted
        # pylint: disable=no-member
        return self.markdown.toc

    @utils.cached_property
    def formatted(self):
        """Markdown formatted content."""
        return self.markdown.convert(
            self.content) if self.content else None


class TextDocumentFormat(DocumentFormat):
    pass


class YamlDocumentFormat(DocumentFormat):

    def _parse_content(self):
        return None

    def _parse_front_matter(self):
        return doc_front_matter.DocumentFrontMatter(
            self._doc, raw_front_matter=self.raw_content)



"""Indexes for tracking the changes to pods across deployments and builds."""

import datetime
import logging
import sys
import traceback
import configparser
import git
import progressbar
import texttable
from multiprocessing import pool
from grow.common import progressbar_non
from grow.common import utils as common_utils
from protorpc import protojson
from . import messages
from . import utils


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class CorruptIndexError(Error):
    pass


class DeploymentError(Error):
    """Errors that occured during the rendering."""

    def __init__(self, message, err, err_tb):
        super(DeploymentError, self).__init__(message)
        self.err = err
        self.err_tb = err_tb


class DeploymentErrors(Error):
    """Errors that occured during the deployment."""

    def __init__(self, message, errors):
        super(DeploymentErrors, self).__init__(message)
        self.errors = errors


class Diff(object):
    POOL_SIZE = 10  # Thread pool size for applying a diff.
    GIT_LOG_MAX = 25

    @classmethod
    def is_empty(cls, diff):
        return not diff.adds and not diff.deletes and not diff.edits

    @classmethod
    def _format_author(cls, author, include_email=True):
        if not author:
            return ''
        author_name = author.name
        author_email = author.email
        if include_email:
            return '{} <{}>'.format(author_name, author_email)
        return author_name

    @classmethod
    def _make_diff_row(cls, color, label, message):
        label = texttable.get_color_string(color, label)
        path = texttable.get_color_string(
            texttable.bcolors.WHITE, message.path)
        formatted_author = cls._format_author(message.deployed_by, True)
        deployed = str(message.deployed).split('.')[0][
            :-3] if message.deployed else ''
        return [label, path, deployed, formatted_author]

    @classmethod
    def pretty_print(cls, diff):
        last_commit = diff.indexes[0].commit
        new_commit = diff.indexes[1].commit
        last_index = diff.indexes[0]
        new_index = diff.indexes[1]

        table = texttable.Texttable(max_width=0)
        table.set_deco(texttable.Texttable.HEADER)
        rows = []
        rows.append(['Action', 'Path', 'Last deployed', 'By'])
        file_rows = []
        for add in diff.adds:
            file_rows.append(cls._make_diff_row(
                texttable.bcolors.GREEN, 'add', add))
        for edit in diff.edits:
            file_rows.append(cls._make_diff_row(
                texttable.bcolors.PURPLE, 'edit', edit))
        for delete in diff.deletes:
            file_rows.append(cls._make_diff_row(
                texttable.bcolors.RED, 'delete', delete))
        file_rows.sort(key=lambda row: row[1])
        rows += file_rows
        table.add_rows(rows)
        logging.info('\n' + table.draw() + '\n')
        if last_index.deployed and last_index.deployed_by:
            logging.info('Last deployed: {} by {}'.format(
                last_index.deployed, cls._format_author(last_index.deployed_by)))
        last_commit_sha = last_commit.sha if last_commit else ''
        new_commit_sha = new_commit.sha if new_commit else ''
        if new_index.deployed_by:
            between_commits = '{}..{}'.format(
                last_commit_sha[:7],
                new_commit_sha[:7])
            if new_commit:
                if new_commit.has_unstaged_changes:
                    between_commits += ' (with unstaged changes)'
            else:
                between_commits += ' (initial commit)'
            logging.info('Diff: {} as {}'.format(
                between_commits, new_index.deployed_by.email))
        if diff.what_changed:
            logging.info(diff.what_changed + '\n')
        if diff.is_partial:
            logging.info(
                '  Note: Partial diffs do not include changes that remove files.\n')

    @classmethod
    def create(cls, index, theirs, repo=None, is_partial=False):
        diff = messages.DiffMessage()
        diff.is_partial = is_partial
        diff.indexes = []
        diff.indexes.append(theirs or messages.IndexMessage())
        diff.indexes.append(index or messages.IndexMessage())

        index_paths_to_shas = {}
        their_paths_to_shas = {}

        for file_message in index.files:
            index_paths_to_shas[file_message.path] = file_message.sha
        for file_message in theirs.files:
            their_paths_to_shas[file_message.path] = file_message.sha

        for path, sha in index_paths_to_shas.items():
            if path in their_paths_to_shas:
                if index_paths_to_shas[path] == their_paths_to_shas[path]:
                    file_message = messages.FileMessage()
                    file_message.path = path
                    file_message.deployed = theirs.deployed
                    file_message.deployed_by = theirs.deployed_by
                    diff.nochanges.append(file_message)
                else:
                    file_message = messages.FileMessage()
                    file_message.path = path
                    file_message.deployed = theirs.deployed
                    file_message.deployed_by = theirs.deployed_by
                    diff.edits.append(file_message)
                del their_paths_to_shas[path]
            else:
                file_message = messages.FileMessage()
                file_message.path = path
                diff.adds.append(file_message)

        # When doing partial diffs we do not have enough information to know
        # which files have been deleted.
        if not is_partial:
            for path, sha in their_paths_to_shas.items():
                file_message = messages.FileMessage()
                file_message.path = path
                file_message.deployed = theirs.deployed
                file_message.deployed_by = theirs.deployed_by
                diff.deletes.append(file_message)

        # What changed in the pod between deploy commits.
        if (repo is not None and index.commit and index.commit.sha and theirs.commit
                and theirs.commit.sha):
            try:
                what_changed = repo.git.log(
                    '--date=short',
                    '--pretty=format:[%h] %ad <%ae> %s',
                    '{}..{}'.format(theirs.commit.sha, index.commit.sha))
                diff.what_changed = what_changed
            except git.GitCommandError:
                logging.info('Unable to determine changes between deploys.')

        # If on the original deploy show commit log messages only.
        elif (repo is not None
              and index.commit and index.commit.sha):
            what_changed = repo.git.log(
                '--date=short',
                '--pretty=format:[%h] %ad <%ae> %s')
            diff.what_changed = what_changed

        return diff

    @classmethod
    def to_string(cls, message):
        return protojson.encode_message(message)

    @classmethod
    def apply(cls, message, paths_to_rendered_doc, write_func, batch_write_func, delete_func,
              threaded=True, batch_writes=False):
        if pool is None:
            text = 'Deployment is unavailable in this environment.'
            raise common_utils.UnavailableError(text)
        apply_errors = []
        thread_pool = None
        if threaded:
            thread_pool = pool.ThreadPool(cls.POOL_SIZE)
        diff = message
        num_files = len(diff.adds) + len(diff.edits) + len(diff.deletes)
        text = 'Deploying: %(value)d/{} (in %(time_elapsed).9s)'
        widgets = [progressbar.FormatLabel(text.format(num_files))]
        progress = progressbar_non.create_progressbar(
            "Deploying...", widgets=widgets, max_value=num_files)

        def run_func(kwargs):
            """Run an arbitrary function."""
            try:
                kwargs['func'](*kwargs['args'])
                return True
            # pylint: disable=broad-except
            except Exception as err:
                _, _, err_tb = sys.exc_info()
                return DeploymentError("Error deploying {}".format(kwargs['args']), err, err_tb)

        if batch_writes:
            writes_paths_to_rendered_doc = {}
            for file_message in diff.adds:
                writes_paths_to_rendered_doc[file_message.path] = \
                    paths_to_rendered_doc[file_message.path]
            for file_message in diff.edits:
                writes_paths_to_rendered_doc[file_message.path] = \
                    paths_to_rendered_doc[file_message.path]
            deletes_paths = [
                file_message.path for file_message in diff.deletes]
            if writes_paths_to_rendered_doc:
                batch_write_func(writes_paths_to_rendered_doc)
            if deletes_paths:
                delete_func(deletes_paths)
        else:
            progress.start()
            threaded_args = []

            for file_message in diff.adds:
                rendered_doc = paths_to_rendered_doc[file_message.path]
                threaded_args.append({
                    'func': write_func,
                    'args': (rendered_doc,),
                })
            for file_message in diff.edits:
                rendered_doc = paths_to_rendered_doc[file_message.path]
                threaded_args.append({
                    'func': write_func,
                    'args': (rendered_doc,),
                })
            for file_message in diff.deletes:
                threaded_args.append({
                    'func': delete_func,
                    'args': (file_message.path,),
                })

            if threaded:
                results = thread_pool.imap_unordered(run_func, threaded_args)
                for result in results:
                    if isinstance(result, Exception):
                        apply_errors.append(result)
                    progress.update(progress.value + 1)
            else:
                for kwargs in threaded_args:
                    try:
                        kwargs['func'](*kwargs['args'])
                        progress.update(progress.value + 1)
                    except DeploymentError as err:
                        apply_errors.append(err)

        if threaded:
            thread_pool.close()
            thread_pool.join()
        if not batch_writes:
            progress.finish()

        if apply_errors:
            for error in apply_errors:
                print(error.message)
                print(error.err)
                traceback.print_tb(error.err_tb)
                print('')
            text = 'There were {} errors during deployment.'
            raise DeploymentErrors(text.format(
                len(apply_errors)), apply_errors)

    @classmethod
    def stream(cls, theirs, content_generator, repo=None, is_partial=False):
        """Render the content and create a diff passing on only the changed content."""
        index = Index.create()
        if repo:
            Index.add_repo(index, repo)
        paths_to_rendered_doc = {}
        diff = messages.DiffMessage()
        diff.is_partial = is_partial
        diff.indexes = []
        diff.indexes.append(theirs or messages.IndexMessage())
        diff.indexes.append(index or messages.IndexMessage())

        index_paths_to_shas = {}
        their_paths_to_shas = {}

        for file_message in theirs.files:
            their_paths_to_shas[file_message.path] = file_message.sha

        for rendered_doc in content_generator:
            path = rendered_doc.path
            index_paths_to_shas[path] = Index.add_file(index, rendered_doc).sha

            if path in their_paths_to_shas:
                if index_paths_to_shas[path] == their_paths_to_shas[path]:
                    file_message = messages.FileMessage()
                    file_message.path = path
                    file_message.deployed = theirs.deployed
                    file_message.deployed_by = theirs.deployed_by
                    diff.nochanges.append(file_message)
                else:
                    file_message = messages.FileMessage()
                    file_message.path = path
                    file_message.deployed = theirs.deployed
                    file_message.deployed_by = theirs.deployed_by
                    diff.edits.append(file_message)
                    paths_to_rendered_doc[path] = rendered_doc
                del their_paths_to_shas[path]
            else:
                file_message = messages.FileMessage()
                file_message.path = path
                diff.adds.append(file_message)
                paths_to_rendered_doc[path] = rendered_doc

        # When doing partial diffs we do not have enough information to know
        # which files have been deleted.
        if not is_partial:
            for path, _ in their_paths_to_shas.items():
                file_message = messages.FileMessage()
                file_message.path = path
                file_message.deployed = theirs.deployed
                file_message.deployed_by = theirs.deployed_by
                diff.deletes.append(file_message)

        # What changed in the pod between deploy commits.
        if (repo is not None and index.commit and index.commit.sha and theirs.commit
                and theirs.commit.sha):
            try:
                what_changed = repo.git.log(
                    '--date=short',
                    '--pretty=format:[%h] %ad <%ae> %s',
                    '{}..{}'.format(theirs.commit.sha, index.commit.sha))
                diff.what_changed = what_changed
            except git.GitCommandError:
                logging.info('Unable to determine changes between deploys.')

        # If on the original deploy show commit log messages only.
        elif (repo is not None
              and index.commit and index.commit.sha):
            what_changed = repo.git.log(
                '--date=short',
                '--pretty=format:[%h] %ad <%ae> %s')
            changed_lines = what_changed.splitlines()
            num_lines = len(changed_lines)
            if num_lines > cls.GIT_LOG_MAX:
                changed_lines = changed_lines[:cls.GIT_LOG_MAX]
                changed_lines.append(
                    ' ... +{} more commits.'.format(num_lines-cls.GIT_LOG_MAX))
                what_changed = '\n'.join(changed_lines)
            diff.what_changed = what_changed

        return diff, index, paths_to_rendered_doc


class Index(object):

    @classmethod
    def create(cls, paths_to_rendered_doc=None):
        message = messages.IndexMessage()
        message.deployed = datetime.datetime.now()
        message.files = []
        if paths_to_rendered_doc is None:
            return message
        for _, rendered_doc in paths_to_rendered_doc.items():
            cls.add_file(message, rendered_doc)
        return message

    @classmethod
    def add_file(cls, message, rendered_doc):
        pod_path = '/' + rendered_doc.path.lstrip('/')
        file_message = messages.FileMessage(
            path=pod_path, sha=rendered_doc.hash)
        message.files.append(file_message)
        return file_message

    @classmethod
    def add_repo(cls, message, repo):
        config = repo.config_reader()
        try:
            message.deployed_by = messages.AuthorMessage(
                name=config.get('user', 'name'),
                email=config.get('user', 'email'))
        except (configparser.NoSectionError, configparser.NoOptionError):
            logging.warning("Couldn't find user info in repository config.")
        try:
            message.commit = utils.create_commit_message(repo)
        except utils.NoGitHeadError as e:
            logging.warning(e)
        return message

    @classmethod
    def to_string(cls, message):
        return protojson.encode_message(message)

    @classmethod
    def from_string(cls, content):
        return protojson.decode_message(messages.IndexMessage, content)



"""
Cache for storing and retrieving data in a pod.

Supports arbitrary data based on a cache key.

The contents of the cache should be raw and not internationalized as it will
be shared between locales.
"""

import re


FILE_OBJECT_CACHE = 'objectcache.json'
FILE_OBJECT_SUB_CACHE = 'objectcache.{}.json'


class ObjectCache(object):
    """Object cache for caching arbitrary data in a pod."""

    def __contains__(self, key):
        return key in self._cache

    def __init__(self):
        self._cache = {}
        self._is_dirty = False
        self.reset()

    def add(self, key, value):
        """Add a new item to the cache or overwrite an existing value."""
        if not self._is_dirty and (key not in self._cache or self._cache[key] != value):
            self._is_dirty = True
        self._cache[key] = value

    def add_all(self, key_to_cached):
        """Update the cache with a preexisting set of data."""
        for key, value in key_to_cached.items():
            self.add(key, value)

    def export(self):
        """Returns the raw cache data."""
        return self._cache

    def get(self, key):
        """Retrieve the value from the cache."""
        return self._cache.get(key, None)

    @property
    def is_dirty(self):
        """Have the contents of the object cache been modified?"""
        return self._is_dirty

    def mark_clean(self):
        """Mark that the object cache is clean."""
        self._is_dirty = False

    def remove(self, key):
        """Removes a single element from the cache."""
        self._is_dirty = True
        return self._cache.pop(key, None)

    def reset(self):
        """Reset the internal cache object."""
        self._cache = {}
        self._is_dirty = False

    def search(self, pattern):
        """Search through the cache and return all the matching elements."""
        if type(pattern) is not type(re.compile('.')):
            pattern = re.compile(pattern)

        results = {}

        for key, value in self._cache.items():
            if pattern.search(key) is not None:
                results[key] = value

        return results



class Error(Exception):
    """Base Exception class for module."""

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class DocumentDoesNotExistError(Error, ValueError):
    """Document path does not exist as a document."""
    pass


class RouteNotFoundError(Error, KeyError):
    """Raised when a resource is not found within the pod."""


class NoPathError(Error, ValueError):
    """Raised when a path is requested for a resource that doesn't have one."""


class PodConfigurationError(Error, ValueError):
    pass


class PodNotFoundError(Error):
    pass


class TestFailedError(Error, AssertionError):
    pass


class BuildError(Error):
    exception = None
    controller = None
    traceback = None


class BadNameError(Error, ValueError):
    pass


class BadYamlError(Error, ValueError):
    pass



"""NPM installer class."""

import subprocess
from grow.sdk import sdk_utils
from grow.sdk.installers import base_installer


INSTALL_HASH_FILE = '/node_modules/install-hash.txt'
NPM_LOCK_FILE = '/package-lock.json'
PACKAGE_FILE = '/package.json'
YARN_LOCK_FILE = '/yarn.lock'


class NpmInstaller(base_installer.BaseInstaller):
    """Grow npm and yarn installer."""

    KIND = 'npm'

    def __init__(self, pod, config):
        super(NpmInstaller, self).__init__(pod, config)
        self._using_yarn = None

    @property
    def post_install_messages(self):
        """List of messages to display after installing."""
        if self.using_yarn:
            return ['Finished: yarn']
        return ['Finished: npm']

    @property
    def should_run(self):
        """Should the installer run?"""
        if not self.pod.file_exists(PACKAGE_FILE):
            return False

        if self.config.get('force', None):
            return True

        if self.pod.file_exists(INSTALL_HASH_FILE):
            existing_hash = self.pod.read_file(INSTALL_HASH_FILE)
            current_hash = self._generate_hashes()
            if existing_hash == current_hash:
                self.pod.logger.info('Skipping npm/yarn install since no changes detected.')
                self.pod.logger.info('   Use `grow install -f` to force install.')
                return False
        return True

    @property
    def using_yarn(self):
        """Is the pod using yarn?"""
        if self._using_yarn is None:
            self._using_yarn = self.pod.file_exists(YARN_LOCK_FILE)
        return self._using_yarn

    def _check_prerequisites_npm(self):
        """Check if required prerequisites are installed or available."""
        status_command = self._nvm_command('npm --version > /dev/null 2>&1')
        not_found = subprocess.call(
            status_command, **self.subprocess_args(shell=True)) == 127
        if not_found:
            install_commands = []
            if sdk_utils.PLATFORM == 'linux':
                install_commands.append(
                    'On Linux, you can install npm using: `apt-get install nodejs`')
            elif sdk_utils.PLATFORM == 'mac':
                install_commands.append(
                    'Using brew (https://brew.sh), you can install using: `brew install node`')
                install_commands.append(('If you do not have brew, you can download '
                                         'Node.js from https://nodejs.org'))
            else:
                install_commands.append(
                    'Download Node.js from https://nodejs.org')
            raise base_installer.MissingPrerequisiteError(
                'The `npm` command was not found.', install_commands=install_commands)

    def _check_prerequisites_yarn(self):
        """Check if required prerequisites are installed or available."""
        status_command = 'yarn --version > /dev/null 2>&1'
        not_found = subprocess.call(
            status_command, **self.subprocess_args(shell=True)) == 127
        if not_found:
            install_commands = [
                'Install yarn from https://yarnpkg.com/en/docs/install']
            raise base_installer.MissingPrerequisiteError(
                'The `yarn` command was not found.', install_commands=install_commands)

    def _generate_hashes(self):
        """Create hash string based on the package and lock files."""
        hash_string = ''
        if self.pod.file_exists(PACKAGE_FILE):
            hash_string = self.pod.hash_file(PACKAGE_FILE)
        if self.pod.file_exists(YARN_LOCK_FILE):
            hash_string = '{} {}'.format(hash_string, self.pod.hash_file(YARN_LOCK_FILE))
        if self.pod.file_exists(NPM_LOCK_FILE):
            hash_string = '{} {}'.format(hash_string, self.pod.hash_file(NPM_LOCK_FILE))
        return hash_string

    def _install_npm(self):
        """Install dependencies using npm."""
        install_command = self._nvm_command('npm install')
        process = subprocess.Popen(install_command, **self.subprocess_args(shell=True))
        code = process.wait()
        if not code:
            self.pod.write_file(INSTALL_HASH_FILE, self._generate_hashes())
            return
        raise base_installer.InstallError(
            'There was an error running `npm install`.')

    def _install_yarn(self):
        """Install dependencies using npm."""
        install_command = 'yarn install'
        process = subprocess.Popen(install_command, **self.subprocess_args(shell=True))
        code = process.wait()
        if not code:
            self.pod.write_file(INSTALL_HASH_FILE, self._generate_hashes())
            return
        raise base_installer.InstallError(
            'There was an error running `yarn install`.')

    def _nvm_command(self, command):
        if self.pod.file_exists('/.nvmrc'):
            # Need to source NVM first to get the nvm command to work.
            return '. $NVM_DIR/nvm.sh && nvm exec {}'.format(command)
        return command

    def check_prerequisites(self):
        """Check if required prerequisites are installed or available."""
        if self.using_yarn:
            self._check_prerequisites_yarn()
        else:
            self._check_prerequisites_npm()

    def install(self):
        """Install dependencies."""
        if self.using_yarn:
            return self._install_yarn()
        return self._install_npm()



"""Webreview deployment destination."""

import logging
import os
import urllib.parse
import webreview
from protorpc import messages
from grow.deployments.destinations import base
from grow.deployments import utils
from grow.common import utils as common_utils
from grow.pods import env
from grow.routing import router


class Config(messages.Message):
    env = messages.MessageField(env.EnvConfig, 1)
    project = messages.StringField(2, required=True)
    name = messages.StringField(3)
    server = messages.StringField(4, required=True)
    secure = messages.BooleanField(5, default=True)
    keep_control_dir = messages.BooleanField(6, default=False)
    remote = messages.StringField(8)
    subdomain = messages.StringField(9)
    subdomain_prefix = messages.StringField(10)
    filters = messages.MessageField(router.FilterConfig, 11, repeated=True)
    tags = messages.StringField(12, repeated=True)


class WebReviewDestination(base.BaseDestination):
    KIND = 'webreview'
    Config = Config
    threaded = True
    batch_writes = True

    def __init__(self, *args, **kwargs):
        super(WebReviewDestination, self).__init__(*args, **kwargs)
        if webreview is None:
            raise common_utils.UnavailableError('WebReview deployments are not available in this environment.')
        if self.config.name:
            print ('WARNING: The "name" parameter for webreview deployments is '
                   'deprecated. Use "subdomain" instead, or use the "grow stage '
                   '--subdomain=<subdomain>" command.')
            self.config.subdomain = self.config.name
        if self.config.remote:
            self.config.server, self.config.project = self.config.remote.split('/', 1)
        if self.config.server.startswith('localhost:'):
            self.config.secure = False
        self._webreview = None

    def __str__(self):
        return self.config.server

    def _print_rpc_traceback(self, err):
        """When there is a rpc error, try to output the traceback."""
        if err.tb:
            self.pod.logger.error('')
            self.pod.logger.error('Original Exception Traceback:')
            for line in err.tb:
                self.pod.logger.error(line)
            self.pod.logger.error('')

    def get_env(self):
        """Returns an environment object based on the config."""
        if self.config.env:
            return env.Env(self.config.env)
        subdomain = self._get_subdomain()
        host = '{}-dot-{}'.format(subdomain, self.config.server)
        config = env.EnvConfig(name='webreview', host=host, scheme='https')
        return env.Env(config)

    @property
    def webreview(self):
        if self._webreview is None:
            api_key = os.getenv('WEBREVIEW_API_KEY')
            subdomain = self._get_subdomain()
            self._webreview = webreview.WebReview(
                project=self.config.project,
                name=subdomain,
                host=self.config.server,
                secure=self.config.secure,
                api_key=api_key)
        return self._webreview

    def _get_branch_token(self):
        # Try getting the branch name from CI environments.
        # https://github.com/grow/grow/issues/903
        branch_name = os.getenv('BRANCH_NAME') or os.getenv('CIRCLE_BRANCH')
        if branch_name:
            return branch_name.split('/')[-1]
        repo = common_utils.get_git_repo(self.pod.root)
        try:
            return repo.active_branch.name.split('/')[-1]
        except TypeError as e:
            # Permit staging from detached branches. Note this will clobber
            # other stages originating from detaching the same branch.
            if 'is a detached symbolic reference' not in str(e):
                raise
            # Extract the sha from the error message.
            sha = str(e).split(' ')[-1].replace("'", '')
            return '{}-d'.format(sha[:10])

    def _get_subdomain(self):
        if self.config.subdomain_prefix and not self.config.subdomain:
            token = self._get_branch_token()
            if token in ['master', 'main']:
                return self.config.subdomain_prefix
            else:
                return self.config.subdomain_prefix + '-{}'.format(token)
        return self.config.subdomain

    def deploy(self, *args, **kwargs):
        repo = kwargs.get('repo')
        if repo:
            subdomain = self._get_subdomain()
            self.webreview.name = subdomain
            try:
                self.webreview.commit = utils.create_commit_message(repo)
            except ValueError:
                raise ValueError(
                    'Cannot deploy to WebReview from a Git repository without a HEAD.'
                    ' Commit first then deploy to WebReview.')
        result = super(WebReviewDestination, self).deploy(*args, **kwargs)
        if self.success:
            finalize_response = self.webreview.finalize()
            if 'fileset' in finalize_response:
                url = finalize_response['fileset']['url']
                # Append the homepage path to the staging link.
                result = urllib.parse.urlparse(url)
                if not result.path and self.pod and self.pod.get_home_doc():
                  home_doc = self.pod.get_home_doc()
                  url = url.rstrip('/') + home_doc.url.path
                logging.info('Staged: %s', url)
        return result

    def login(self, account='default', reauth=False):
        self.webreview.login(account, reauth=reauth)

    def prelaunch(self, dry_run=False):
        super(WebReviewDestination, self).prelaunch(dry_run=dry_run)

    def postlaunch(self, dry_run=False):
        super(WebReviewDestination, self).postlaunch(dry_run=dry_run)

    def test(self):
        # Skip the default "can write files at destination" test.
        pass

    def read_file(self, path):
        try:
            paths_to_contents, errors = self.webreview.read([path])
            if path not in paths_to_contents:
                raise IOError('{} not found.'.format(path))
            if errors:
                raise base.Error(errors)
            return paths_to_contents[path]
        except webreview.RpcError as e:
            self._print_rpc_traceback(e)
            raise base.Error(e.message)

    def write_files(self, paths_to_rendered_doc):
        try:
            paths_to_rendered_doc, errors = self.webreview.write(paths_to_rendered_doc)
            if errors:
                raise base.Error(errors)
            return paths_to_rendered_doc
        except webreview.RpcError as e:
            self._print_rpc_traceback(e)
            raise base.Error(e.message)

    def delete_file(self, paths):
        try:
            paths, errors = self.webreview.delete(paths)
            if errors:
                raise base.Error(errors)
            return paths
        except webreview.RpcError as e:
            self._print_rpc_traceback(e)
            # NOTE: Sometimes, a file may fail to delete. This usually happens
            # because the uploaded fileset is in an inconsistent state. The
            # inconsistency may happen due to uploading the manifest, and then
            # subsequently canceling the upload. Subsequent uploads will
            # attempt to delete files that may no longer exist. It's acceptable
            # to not fail outright when deletions fail during staging, so avoid
            # raising the error.


# Support legacy "jetway" destination. Remove this in a future release.
class LegacyJetwayDestination(WebReviewDestination):
    KIND = 'jetway'



"""Grow static document."""

import os
import re
from grow.common import urls
from grow.routing import path_filter as grow_path_filter


# SHA 256 hash length: 64 characters.
FINGERPRINT_RE = re.compile(
    r'(.*)(-[a-f0-9]{64})((\.min|)[\.][a-z0-9]{1,5})$', re.IGNORECASE)


class StaticDocument(object):
    """Static document."""

    def __eq__(self, other):
        return (self.pod == other.pod and self.pod_path == other.pod_path
                and other.locale == self.locale)

    def __init__(self, pod, pod_path, locale=None):
        self.pod = pod
        self.pod_path = pod_path
        self.locale = locale
        self.config = self.pod.router.get_static_config_for_pod_path(pod_path)

        # When localized the base string is changed.
        self._base_path_format = self.config['serve_at']
        self._base_source_path = self.source_path
        self._base_source_path_index = self.source_paths.index(
            self._base_source_path)

        self.use_locale = locale is not None and locale != pod.podspec.default_locale
        if self.use_locale and 'localization' in self.config:
            inherited = {
                'fingerprinted': self.config.get('fingerprinted', False),
            }
            self.config = self.config.get('localization')
            self.config.update(inherited)

        self.use_fallback = 'fallback' not in self.config or self.config['fallback']

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        if self.locale:
            return "<StaticDocument({}, locale='{}')>".format(
                self.pod_path, self.locale)
        return "<StaticDocument({})>".format(self.pod_path)

    @staticmethod
    def strip_fingerprint(serving_path):
        """Strip the fingerprint off a serving path."""
        return FINGERPRINT_RE.sub(r'\1\3', serving_path)

    @property
    def base_path_format(self):
        """Base (non-localized) path format for the static document."""
        return '{}{}'.format(
            self._base_path_format,
            self.sub_base_pod_path)

    @property
    def exists(self):
        """Does the static file exist in the pod?"""
        return self.pod.file_exists(self.source_pod_path)

    @property
    def filter(self):
        """Static configuration filter."""
        return self.config.get('filter', {})

    @property
    def modified(self):
        """File modified timestamp."""
        return self.pod.file_modified(self.pod_path)

    @property
    def size(self):
        """Static file size."""
        return self.pod.file_size(self.pod_path)

    @property
    def fingerprint(self):
        """Fingerprint of the file contents."""
        return self.pod.hash_file(self.pod_path)

    @property
    def fingerprinted(self):
        """Fingerprint file contents?"""
        return self.config.get('fingerprinted', False)

    @property
    def path_filter(self):
        """Path filter for the static document."""
        static_filter = self.filter
        if static_filter:
            return grow_path_filter.PathFilter(
                static_filter.get('ignore_paths'), static_filter.get('include_paths'))
        return self.pod.path_filter

    @property
    def path_format(self):
        """Path format for the static document."""
        return '{}{}'.format(
            self.config['serve_at'],
            self.sub_base_pod_path)

    @property
    def source_format(self):
        """Path format for the source of the document."""
        return '{}{}'.format(
            self.source_path,
            self.sub_base_pod_path)

    @property
    def source_formats(self):
        """Path formats for the source of the document."""
        formats = []
        for source_path in self.source_paths:
            formats.append('{}{}'.format(
                source_path,
                self.sub_base_pod_path))
        return formats

    @property
    def serving_path(self):
        """Serving path for the static document."""
        if self.source_pod_path != self.pod_path or not self.use_fallback:
            path = self.pod.path_format.format_static(
                self.path_format, locale=self.locale,
                fingerprint=self.fingerprint)
        else:
            # Fall back to use the default locale for the formatted path.
            path = self.pod.path_format.format_static(
                self.base_path_format, locale=self.pod.podspec.default_locale,
                fingerprint=self.fingerprint)

        if not self.fingerprinted:
            return path

        base, ext = os.path.splitext(path)
        # Special case to preserve ".min.<ext>" extensions.
        if base.endswith('.min'):
            base = base[:-4]
            return '{}-{}.min{}'.format(base, self.fingerprint, ext)
        return '{}-{}{}'.format(base, self.fingerprint, ext)

    @property
    def serving_path_parameterized(self):
        """Parameterized serving path for the static document."""
        return self.pod.path_format.format_static(
            self.path_format, locale=self.locale, parameterize=True)

    @property
    def source_path(self):
        """Source path for the static document or default source path to first available."""
        path = self.config.get('static_dir')
        if path:
            return path
        for source_path in self.source_paths:
            if self.pod_path.startswith(source_path):
                return source_path
        # Default to the same index as the base source path for localized paths.
        return self.source_paths[self._base_source_path_index or 0]

    @property
    def source_paths(self):
        """Source paths for the static document."""
        paths = self.config.get('static_dirs')
        if paths:
            return paths
        return [self.config.get('static_dir')]

    @property
    def source_pod_path(self):
        """Source path for the static document with missing file fallback."""
        source_path = self.pod.path_format.format_static(
            self.source_format, locale=self.locale)
        # Fall back to the pod path if using locale and the localized
        # version does not exist.
        if self.use_locale and self.use_fallback and not self.pod.file_exists(source_path):
            source_path = self.pod_path
        return source_path

    @property
    def sub_base_pod_path(self):
        """Unique portion of the base static file."""
        return self.pod_path[len(self._base_source_path):]

    @property
    def sub_pod_path(self):
        """Unique portion of the full static file."""
        return self.pod_path[len(self.source_path):]

    @property
    def url(self):
        """URL of the file contents."""
        return urls.Url(
            path=self.serving_path,
            host=self.pod.env.host,
            port=self.pod.env.port,
            scheme=self.pod.env.scheme)



"""Google Sheets for translating pod content."""

import datetime
import io
import progressbar
import random
from babel.messages import catalog
from babel.messages import pofile
from googleapiclient import errors
from grow.common import progressbar_non
from grow.preprocessors import google_drive
from grow.translators import errors as translator_errors
try:
    import io as StringIO
except ImportError:  # pragma: no cover
    try:
        import io
    except ImportError:
        from io import StringIO
from . import base


class AccessLevel(object):
    COMMENTER = 'commenter'
    OWNER = 'owner'
    READER = 'reader'
    WRITER = 'writer'


DEFAULT_ACCESS_LEVEL = AccessLevel.WRITER
OAUTH_SCOPE = 'https://www.googleapis.com/auth/spreadsheets'


class GoogleSheetsTranslator(base.Translator):
    COLOR_DEEP_PURPLE_50 = {
        'red': 0.929,
        'green': 0.905,
        'blue': 0.964,
    }
    COLOR_GREY_200 = {
        'red': .933,
        'blue': .933,
        'green': .933,
    }
    COLOR_GREY_500 = {
        'red': .6196,
        'blue': .6196,
        'green': .6196,
    }
    KIND = 'google_sheets'
    HEADER_ROW_COUNT = 1
    # Source locale, translation locale, message location.
    HEADER_LABELS = [None, None, 'Extracted comments', 'Reference']
    SHEETS_BASE_URL = 'https://docs.google.com/spreadsheets/d/'
    has_immutable_translation_resources = False
    has_multiple_langs_in_one_resource = True

    def __init__(self, pod, config=None, project_title=None, instructions=None):
        super(GoogleSheetsTranslator, self).__init__(
            pod, config=config, project_title=project_title,
            instructions=instructions)
        self.update_meta_after_upload = False

    def needs_meta_update(self):
        """Allow to be flagged for additional meta update after uploading."""
        return self.update_meta_after_upload

    def _catalog_has_comments(self, catalog):
        for message in catalog:
            if not message.id:
                continue

            if message.auto_comments:
                return True
        return False

    def _content_hash(self, location, locale):
        return hash((location, locale)) % (10 ** 8)  # 10 Digits of the hash.

    def _create_service(self):  # pragma: no cover
        return google_drive.BaseGooglePreprocessor.create_service(
            'sheets', 'v4')

    def _download_sheet(self, spreadsheet_id, locale):
        service = self._create_service()
        rangeName = "'{}'!A:D".format(locale)
        try:
            # pylint: disable=no-member
            resp = service.spreadsheets().values().get(
                spreadsheetId=spreadsheet_id, range=rangeName).execute()
        except errors.HttpError as e:
            if e.resp['status'] == '400':
                raise translator_errors.NotFoundError(
                    'Translation for {} not found.'.format(locale))
            raise

        # Check for spreadsheets that are missing columns.
        column_count = len(self.HEADER_LABELS)
        if len(resp['values'][0]) < column_count:
            missing_columns = [None] * (column_count - len(resp['values'][0]))
            resp['values'][:] = [i + missing_columns for i in resp['values']]
        return resp['values']

    def _download_sheets(self, spreadsheet_id, locales):
        service = self._create_service()

        # Find out which locales have sheets.
        # pylint: disable=no-member
        resp = service.spreadsheets().get(
            spreadsheetId=spreadsheet_id).execute()
        existing_locales = []
        for sheet in resp['sheets']:
            existing_locales.append(sheet['properties']['title'])

        rangeNames = []
        rangeLocales = []
        skipped_locales = []
        for locale in locales:
            # Skip locales that do not have sheets.
            if locale not in existing_locales:
                skipped_locales.append(locale)
                continue

            rangeLocales.append(locale)
            rangeNames.append("'{}'!A:D".format(locale))

        if skipped_locales:
            self.pod.logger.info(
                'Skipping locales, missing sheet: {}'.format(', '.join(skipped_locales)))

        try:
            # pylint: disable=no-member
            batch_resp = service.spreadsheets().values().batchGet(
                spreadsheetId=spreadsheet_id, ranges=rangeNames).execute()

            locale_to_values = {}
            for i, locale in enumerate(rangeLocales):
                resp = batch_resp['valueRanges'][i]

                # Check for spreadsheets that are missing columns.
                column_count = len(self.HEADER_LABELS)
                if len(resp['values'][0]) < column_count:
                    missing_columns = [None] * (column_count - len(resp['values'][0]))
                    resp['values'][:] = [i + missing_columns for i in resp['values']]
                locale_to_values[locale] = resp['values']

            return locale_to_values
        except errors.HttpError as e:
            if e.resp['status'] == '400':
                raise translator_errors.NotFoundError(
                    'Translation for {} not found.'.format(locales))
            raise

    def _download_content(self, stat, values):
        # Ignore the header row.
        values.pop(0)
        babel_catalog = catalog.Catalog(stat.lang)
        for row in values:
            # Skip empty rows.
            if not row or self._is_empty_row(row):
                continue
            source = row[0]
            translation = row[1] if len(row) > 1 else None
            babel_catalog.add(source, translation, auto_comments=[],
                              context=None, flags=[])
        updated_stat = base.TranslatorStat(
            url=stat.url,
            lang=stat.lang,
            downloaded=datetime.datetime.now(),
            source_lang=stat.source_lang,
            ident=stat.ident)
        fp = io.BytesIO()
        pofile.write_po(fp, babel_catalog)
        fp.seek(0)
        content = fp.read()
        return updated_stat, content

    def _is_empty_row(self, row):
        return bool(set(row) - set((None, ''))) is False

    def _upload_catalogs(self, catalogs, source_lang, prune=False):
        if not catalogs:
            raise translator_errors.NoCatalogsError('Missing catalogs to upload.')

        project_title = self.project_title
        source_lang = str(source_lang)
        locales_to_sheet_ids = {}

        # Get existing sheet ID (if it exists) from one stat.
        spreadsheet_id = None
        stats_to_download = self._get_stats_to_download([])
        if stats_to_download:
            stat = list(stats_to_download.values())[0]
            spreadsheet_id = stat.ident if stat else None

        # NOTE: Manging locales across multiple spreadsheets is unsupported.
        service = self._create_service()
        if spreadsheet_id:
            # pylint: disable=no-member
            resp = service.spreadsheets().get(
                spreadsheetId=spreadsheet_id).execute()
            for sheet in resp['sheets']:
                locales_to_sheet_ids[sheet['properties']['title']] = \
                    sheet['properties']['sheetId']
            catalogs_to_create = []
            sheet_ids_to_catalogs = {}

            for catalog in catalogs:
                existing_sheet_id = locales_to_sheet_ids.get(
                    str(catalog.locale))
                if existing_sheet_id:
                    sheet_ids_to_catalogs[existing_sheet_id] = catalog
                elif source_lang != str(catalog.locale):
                    catalogs_to_create.append(catalog)

            requests = []

            if catalogs_to_create:
                requests += self._generate_create_sheets_requests(
                    catalogs_to_create, source_lang)

            if sheet_ids_to_catalogs:
                requests += self._generate_update_sheets_requests(
                    sheet_ids_to_catalogs, source_lang, spreadsheet_id,
                    prune=prune)

            self._perform_batch_update(spreadsheet_id, requests)
        else:
            # Create a new spreadsheet and use the id.
            service = self._create_service()
            # pylint: disable=no-member
            resp = service.spreadsheets().create(body={
                'properties': {
                    'title': project_title,
                },
            }).execute()
            spreadsheet_id = resp['spreadsheetId']

            requests = []
            requests += self._generate_create_sheets_requests(
                catalogs, source_lang)
            self._perform_batch_update(spreadsheet_id, requests)

            if 'acl' in self.config:
                self._do_update_acl(spreadsheet_id, self.config['acl'])

        stats = []
        for catalog in catalogs:
            if str(catalog.locale) == source_lang:
                continue
            url = '{}{}'.format(self.SHEETS_BASE_URL, spreadsheet_id)
            lang = str(catalog.locale)
            if lang in locales_to_sheet_ids:
                url += '#gid={}'.format(locales_to_sheet_ids[lang])
            stat = base.TranslatorStat(
                url=url,
                lang=lang,
                source_lang=source_lang,
                uploaded=datetime.datetime.now(),
                ident=spreadsheet_id)
            stats.append(stat)
        return stats

    def _create_header_row_data(self, source_lang, lang):
        return {
            'values': [
                {'userEnteredValue': {'stringValue': source_lang}},
                {'userEnteredValue': {'stringValue': lang}},
                {'userEnteredValue': {'stringValue': self.HEADER_LABELS[2]}},
                {'userEnteredValue': {'stringValue': self.HEADER_LABELS[3]}},
            ]
        }

    def _create_catalog_row(self, id, value, comments, locations):
        comments = [] if comments is None else comments
        return {
            'values': [
                {'userEnteredValue': {'stringValue': id}},
                {'userEnteredValue': {'stringValue': value}},
                {'userEnteredValue': {'stringValue': '\n'.join(comments)}},
                {'userEnteredValue': {
                    'stringValue': ', '.join(t[0] for t in locations)}},
            ],
        }

    def _create_catalog_rows(self, catalog, prune=False):
        rows = []
        for message in catalog:
            if not message.id:
                continue

            if prune and not message.locations:
                continue

            rows.append(self._create_catalog_row(
                message.id, message.string, message.auto_comments, message.locations))
        return rows

    def _diff_data(self, existing_values, catalog):
        existing_rows = []
        new_rows = []
        removed_rows = []

        for value in existing_values:
            if not value or self._is_empty_row(value):  # Skip empty rows.
                continue
            existing_rows.append({
                'source': value[0],
                'translation': value[1] if len(value) > 1 else None,
                'comments': value[2] if len(value) > 2 else [],
                'locations': value[3] if len(value) > 3 else [],
                'updated': False,  # Has changed from the downloaded value.
                'matched': False,  # Has been matched to the downloaded values.
            })

        for message in catalog:
            if not message.id:
                continue

            found = False

            # Update for any existing values.
            for value in existing_rows:
                if value['source'] == message.id:
                    value['updated'] = (
                        value['translation'] != message.string or
                        value['locations'] != message.locations)
                    value['translation'] = message.string
                    value['locations'] = message.locations
                    value['comments'] = message.auto_comments
                    value['matched'] = True
                    found = True
                    break

            if found == True:
                continue

            new_rows.append({
                'source': message.id,
                'translation': message.string,
                'locations': message.locations,
                'comments': message.auto_comments,
            })

        for index, value in enumerate(existing_rows):
            if not value['matched'] or len(value['locations']) == 0:
                removed_rows.append(index)

            # Reset the locations when not found in catalog.
            if not value['matched']:
                value['locations'] = []

        return (existing_rows, new_rows, removed_rows)

    def _generate_comments_column_requests(self, sheet_id, catalog):
        requests = []
        sheet_range = {
            'sheetId': sheet_id,
            'dimension': 'COLUMNS',
            'startIndex': 2,
            'endIndex': 3,
        }

        if self._catalog_has_comments(catalog):
            requests.append({
                'updateDimensionProperties': {
                    'range': sheet_range,
                    'properties': {
                        'hiddenByUser': False,
                    },
                    'fields': 'hiddenByUser',
                },
            })
        else:
            requests.append({
                'updateDimensionProperties': {
                    'range': sheet_range,
                    'properties': {
                        'hiddenByUser': True,
                    },
                    'fields': 'hiddenByUser',
                },
            })

        return requests

    def _generate_create_sheets_requests(self, catalogs, source_lang):
        # Create sheets.
        requests = []
        for catalog in catalogs:
            sheet_id = self._generate_new_sheet_id()
            lang = str(catalog.locale)
            # Create a new sheet.
            requests.append({
                'addSheet': {
                    'properties': {
                        'sheetId': sheet_id,
                        'title': lang,
                        'gridProperties': {
                            'columnCount': 4,
                            'rowCount': self.HEADER_ROW_COUNT + 1,
                            'frozenRowCount': 1,
                            'frozenColumnCount': 1,
                        },
                    },
                },
            })

            # Add the data to the new sheet.
            _, new_rows, _ = self._diff_data([], catalog)
            row_data = []
            row_data.append(self._create_header_row_data(source_lang, lang))

            if len(new_rows):
                for value in new_rows:
                    row_data.append(self._create_catalog_row(
                        value['source'], value['translation'],
                        value['comments'], value['locations']))

                requests.append({
                    'appendCells': {
                        'sheetId': sheet_id,
                        'fields': 'userEnteredValue',
                        'rows': row_data,
                    },
                })

            # Format the new sheet.
            requests += self._generate_style_requests(
                sheet_id, catalog=catalog)

            # Size the initial sheet columns.
            # Not part of the style request to respect user's choice to resize.
            requests.append({
                'updateDimensionProperties': {
                    'range': {
                        'sheetId': sheet_id,
                        'dimension': 'COLUMNS',
                        'startIndex': 0,
                        'endIndex': 3,
                    },
                    'properties': {
                        'pixelSize': 400,  # Source and Translation Columns
                    },
                    'fields': 'pixelSize',
                },
            })

            requests.append({
                'updateDimensionProperties': {
                    'range': {
                        'sheetId': sheet_id,
                        'dimension': 'COLUMNS',
                        'startIndex': 3,
                        'endIndex': 4,
                    },
                    'properties': {
                        'pixelSize': 200,  # Location Column
                    },
                    'fields': 'pixelSize',
                },
            })

            requests += self._generate_comments_column_requests(
                sheet_id, catalog)

        return requests

    def _generate_filter_view_requests(self, sheet_id, sheet, filter_view):
        requests = []

        is_filtered = False
        if sheet and 'filterViews' in sheet:
            for existing_range in sheet['filterViews']:
                if existing_range['filterViewId'] == filter_view['filterViewId']:
                    is_filtered = True
                    requests.append({
                        'updateFilterView': {
                            'filter': filter_view,
                            'fields': 'range,title,criteria',
                        },
                    })
                    break

        if not is_filtered:
            requests.append({
                'addFilterView': {
                    'filter': filter_view,
                },
            })

        return requests

    def _generate_new_sheet_id(self):
        return random.randrange(100, 9999999)

    def _generate_style_requests(self, sheet_id, sheet=None, catalog=None):
        formats = {}

        formats['header_cell'] = {
            'backgroundColor': self.COLOR_GREY_200,
            'textFormat': {
                'bold': True
            },
        }

        formats['info_cell'] = {
            'wrapStrategy': 'WRAP',
            'textFormat': {
                'foregroundColor': self.COLOR_GREY_500,
            },
        }

        formats['missing_cell'] = {
            'backgroundColor': self.COLOR_DEEP_PURPLE_50,
        }

        formats['wrap'] = {'wrapStrategy': 'WRAP'}

        requests = []

        # TODO Figure out how to be smarter about matching conditional formatting.
        # Remove all existing conditional formatting. :(
        if sheet and 'conditionalFormats' in sheet:
            for _ in sheet['conditionalFormats']:
                requests.append({
                    'deleteConditionalFormatRule': {
                        'sheetId': sheet_id,
                        'index': 0
                    }
                })

        # Style header cells.
        requests.append({
            'repeatCell': {
                'fields': 'userEnteredFormat',
                'range': {
                    'sheetId': sheet_id,
                    'startColumnIndex': 0,
                    'startRowIndex': 0,
                    'endRowIndex': self.HEADER_ROW_COUNT,
                },
                'cell': {
                    'userEnteredFormat': formats['header_cell'],
                },
            },
        })

        # Allow the translations and comments to wrap.
        requests.append({
            'repeatCell': {
                'fields': 'userEnteredFormat',
                'range': {
                    'sheetId': sheet_id,
                    'startColumnIndex': 0,
                    'endColumnIndex': 3,
                    'startRowIndex': self.HEADER_ROW_COUNT,
                },
                'cell': {
                    'userEnteredFormat': formats['wrap'],
                },
            },
        })

        # Comment and source cells are muted in styling.
        requests.append({
            'repeatCell': {
                'fields': 'userEnteredFormat',
                'range': {
                    'sheetId': sheet_id,
                    'startColumnIndex': 2,
                    'endColumnIndex': 4,
                    'startRowIndex': self.HEADER_ROW_COUNT,
                },
                'cell': {
                    'userEnteredFormat': formats['info_cell'],
                },
            },
        })

        # Highlight missing translations.
        requests.append({
            'addConditionalFormatRule': {
                'rule': {
                    'ranges': [{
                        'sheetId': sheet_id,
                        'startColumnIndex': 1,
                        'endColumnIndex': 2,
                        'startRowIndex': self.HEADER_ROW_COUNT,
                    }],
                    'booleanRule': {
                        'condition': {'type': 'BLANK'},
                        'format': formats['missing_cell']
                    }
                },
                'index': 0
            }
        })

        # Protect the original values.
        requests += self._generate_style_protected_requests(sheet_id, sheet, {
            'protectedRangeId': sheet_id + 1000001,  # Keep it predictble.
            'range': {
                'sheetId': sheet_id,
                'startColumnIndex': 0,
                'endColumnIndex': 1,
                'startRowIndex': self.HEADER_ROW_COUNT,
            },
            'description': 'Original strings can only be edited in the source files.',
            'warningOnly': True,
        })

        # Protect the comment values.
        requests += self._generate_style_protected_requests(sheet_id, sheet, {
            'protectedRangeId': sheet_id + 1000002,  # Keep it predictble.
            'range': {
                'sheetId': sheet_id,
                'startColumnIndex': 2,
                'endColumnIndex': 3,
                'startRowIndex': self.HEADER_ROW_COUNT,
            },
            'description': 'Comment strings can only be edited in the source files.',
            'warningOnly': True,
        })

        # Protect the location values.
        requests += self._generate_style_protected_requests(sheet_id, sheet, {
            'protectedRangeId': sheet_id + 1000003,  # Keep it predictble.
            'range': {
                'sheetId': sheet_id,
                'startColumnIndex': 3,
                'endColumnIndex': 4,
                'startRowIndex': self.HEADER_ROW_COUNT,
            },
            'description': 'Source strings can only be edited in the source files.',
            'warningOnly': True,
        })

        # Filter view for untranslated strings.
        requests += self._generate_filter_view_requests(sheet_id, sheet, {
            'filterViewId': sheet_id + 3300001,  # Keep it predictble.
            'range': {
                'sheetId': sheet_id,
                'startColumnIndex': 0,
                'endColumnIndex': 4,
                'startRowIndex': 0,
            },
            'title': 'Untranslated Strings',
            'criteria': {
                '1': {
                    'condition': {'type': 'BLANK'}
                },
            },
        })

        return requests

    def _generate_style_protected_requests(self, sheet_id, sheet, protected_range):
        requests = []

        is_protected = False
        if sheet and 'protectedRanges' in sheet:
            for existing_range in sheet['protectedRanges']:
                if existing_range['protectedRangeId'] == protected_range['protectedRangeId']:
                    is_protected = True
                    requests.append({
                        'updateProtectedRange': {
                            'protectedRange': protected_range,
                            'fields': 'range,description,warningOnly',
                        },
                    })
                    break

        if not is_protected:
            requests.append({
                'addProtectedRange': {
                    'protectedRange': protected_range,
                },
            })

        return requests

    def _generate_update_sheets_requests(self, sheet_ids_to_catalogs,
                                         source_lang, spreadsheet_id, prune=False):
        requests = []
        for sheet_id, catalog in sheet_ids_to_catalogs.items():
            lang = str(catalog.locale)
            existing_values = self._download_sheet(spreadsheet_id, lang)
            for x in range(self.HEADER_ROW_COUNT):
                existing_values.pop(0)  # Remove header rows.
            for value in existing_values:
                if not value or self._is_empty_row(value):  # Skip empty rows.
                    continue
                if value not in catalog:
                    source = value[0]
                    translation = value[1] if len(value) > 1 else None
                    catalog.add(source, translation, auto_comments=[],
                                context=None, flags=[])

            # Check for missing columns.
            num_missing_columns = 0
            for column in existing_values[0]:
                if column == None:
                    num_missing_columns += 1

            if num_missing_columns:
                requests.append({
                    'appendDimension': {
                        'sheetId': sheet_id,
                        'dimension': 'COLUMNS',
                        'length': num_missing_columns,
                    },
                })

                # Update the column headers.
                requests.append({
                    'updateCells': {
                        'fields': 'userEnteredValue',
                        'start': {
                            'sheetId': sheet_id,
                            'rowIndex': 0,
                            'columnIndex': 0,
                        },
                        'rows': [
                            self._create_header_row_data(source_lang, lang)
                        ],
                    },
                })

            # Perform a diff of the existing data to what the catalog provides
            # to make targeted changes to the spreadsheet and preserve meta
            # information--such as comments.
            existing_rows, new_rows, removed_rows = self._diff_data(
                existing_values, catalog)

            # Update the existing values in place.
            if len(existing_rows):
                row_data = []
                for value in existing_rows:
                    if not value:  # Skip empty rows.
                        continue
                    row_data.append(self._create_catalog_row(
                        value['source'], value['translation'],
                        value['comments'], value['locations']))
                # NOTE This is not (yet) smart enough to only update small sections
                # with the updated information. Hint: Use value['updated'].
                requests.append({
                    'updateCells': {
                        'fields': 'userEnteredValue',
                        'start': {
                            'sheetId': sheet_id,
                            # Skip header row.
                            'rowIndex': self.HEADER_ROW_COUNT,
                            'columnIndex': 0,
                        },
                        'rows': row_data,
                    },
                })

            # Append new values to end of sheet.
            if len(new_rows):
                # Mark to update the sheet metadata after done.
                self.update_meta_after_upload = True
                row_data = []
                for value in new_rows:
                    row_data.append(self._create_catalog_row(
                        value['source'], value['translation'],
                        value['comments'], value['locations']))

                requests.append({
                    'appendCells': {
                        'sheetId': sheet_id,
                        'fields': 'userEnteredValue',
                        'rows': row_data,
                    },
                })

            # Remove obsolete rows if not included.
            if prune and len(removed_rows):
                for value in reversed(removed_rows):  # Start from the bottom.
                    # NOTE this is ineffecient since it does not combine ranges.
                    # ex: 1, 2, 3 are three requests instead of one request 1-3
                    requests.append({
                        'deleteDimension': {
                            'range': {
                                'sheetId': sheet_id,
                                'dimension': 'ROWS',
                                'startIndex': self.HEADER_ROW_COUNT + value,
                                'endIndex': self.HEADER_ROW_COUNT + value + 1,
                            },
                        },
                    })

            # Sort all rows.
            requests.append({
                'sortRange': {
                    'range': {
                        'sheetId': sheet_id,
                        'startColumnIndex': 0,
                        'startRowIndex': self.HEADER_ROW_COUNT,
                    },
                    'sortSpecs': [
                        {
                            'dimensionIndex': 0,
                            'sortOrder': 'ASCENDING',
                        }
                    ],
                },
            })

            requests += self._generate_comments_column_requests(
                sheet_id, catalog)
        return requests

    def _do_update_acl(self, spreadsheet_id, acl):
        service = google_drive.BaseGooglePreprocessor.create_service(
            'drive', 'v3')
        for item in acl:
            permission = {
                'role': item.get('role', DEFAULT_ACCESS_LEVEL).lower(),
            }
            if 'domain' in item:
                permission['type'] = 'domain'
                permission['domain'] = item['domain']
            elif 'user' in item:
                permission['type'] = 'user'
                permission['emailAddress'] = item['user']
            elif 'group' in item:
                permission['type'] = 'group'
                permission['emailAddress'] = item['group']
            # pylint: disable=no-member
            resp = service.permissions().create(
                fileId=spreadsheet_id,
                body=permission).execute()

    def _perform_batch_update(self, spreadsheet_id, requests):
        service = self._create_service()
        body = {'requests': requests}
        # pylint: disable=no-member
        return service.spreadsheets().batchUpdate(
            spreadsheetId=spreadsheet_id, body=body).execute()

    def _update_acls(self, stats, locales):
        if 'acl' not in self.config:
            return
        stat = stats.values()[0]
        spreadsheet_id = stat.ident
        acl = self.config['acl']
        if not acl:
            return
        self._do_update_acl(spreadsheet_id, acl)

    def _update_meta(self, stat, locale, catalog):
        spreadsheet_id = stat.ident
        service = self._create_service()
        # pylint: disable=no-member
        resp = service.spreadsheets().get(
            spreadsheetId=stat.ident).execute()
        requests = []
        for sheet in resp['sheets']:
            if sheet['properties']['title'] != locale:
                continue
            sheet_id = sheet['properties']['sheetId']
            requests += self._generate_style_requests(
                sheet_id, sheet=sheet, catalog=catalog)
        self._perform_batch_update(spreadsheet_id, requests)

    def download(self, locales, save_stats=True, include_obsolete=False):
        """Override base download to remove the threading and simplify."""
        if not self.pod.file_exists(base.Translator.TRANSLATOR_STATS_PATH):
            text = 'File {} not found. Nothing to download.'
            self.pod.logger.info(text.format(base.Translator.TRANSLATOR_STATS_PATH))
            return
        stats_to_download = self._get_stats_to_download(locales)
        if not stats_to_download:
            return
        num_files = len(stats_to_download)
        text = 'Downloading translations: %(value)d/{} (in %(time_elapsed).9s)'
        widgets = [progressbar.FormatLabel(text.format(num_files))]
        bar = progressbar_non.create_progressbar(
            "Downloading translations...", widgets=widgets, max_value=num_files)
        bar.start()

        spreadsheet_id_to_locales = {}
        for locale in stats_to_download:

            stat = stats_to_download[locale]
            if stat.ident not in spreadsheet_id_to_locales:
                spreadsheet_id_to_locales[stat.ident] = set()
            spreadsheet_id_to_locales[stat.ident].add(locale)

        langs_to_translations = {}
        new_stats = []

        # Batch download for each sheet id.
        for spreadsheet_id in spreadsheet_id_to_locales:
            locales = spreadsheet_id_to_locales[spreadsheet_id]
            locale_to_values = self._download_sheets(spreadsheet_id, locales)

            for i, (lang, stat) in enumerate(stats_to_download.items()):
                if lang not in locale_to_values:
                    continue

                new_stat, content = self._download_content(stat, locale_to_values[lang])
                bar.update(bar.value + 1)
                new_stat.uploaded = stat.uploaded  # Preserve uploaded field.
                langs_to_translations[lang] = content
                new_stats.append(new_stat)

        bar.finish()

        text = 'Importing translations: %(value)d/{} (in %(time_elapsed).9s)'
        widgets = [progressbar.FormatLabel(text.format(num_files))]
        bar = progressbar_non.create_progressbar(
            "Importing translations...", widgets=widgets, max_value=num_files)
        bar.start()

        has_changed_content = False
        unchanged_locales = []
        changed_locales = {}
        for lang, translations in langs_to_translations.items():
            has_changed_content, imported_translations, total_translations = self.pod.catalogs.import_translations(
                    locale=lang, content=translations,
                    include_obsolete=include_obsolete)
            bar.update(bar.value + 1)
            if imported_translations == 0:
                unchanged_locales.append(lang)
            else:
                changed_locales[lang] = {
                    'imported': imported_translations,
                    'total': total_translations,
                }
            if has_changed_content:
                has_changed_content = True

        bar.finish()

        if save_stats and has_changed_content:
            self.save_stats(new_stats)
        self._log_catalog_changes(unchanged_locales, changed_locales)
        return new_stats

    def get_edit_url(self, doc):
        if not doc.locale:
            return
        stats = self._get_stats_to_download([doc.locale])
        if doc.locale not in stats:
            return
        stat = stats[doc.locale]
        return '{}&fvid={}'.format(stat.url, self._content_hash(doc.pod_path, doc.locale))



"""Subommand for uploading translations to a remote source."""

import os
import click
from grow.commands import shared
from grow.commands.subcommands.translations_extract import validate_locales
from grow.common import rc_config
from grow.pods import pods
from grow import storage


CFG = rc_config.RC_CONFIG.prefixed('grow.translations.upload')


@click.command(name='upload')
@shared.pod_path_argument
@click.option('--force/--noforce', '-f', default=CFG.get('force', False), is_flag=True,
              help='Whether to skip the prompt prior to uploading.')
@click.option('--update-acl', default=False, is_flag=True,
              help='Whether to update the ACL on uploaded resources'
                   ' instead of uploading new translation files.')
@click.option('--download/--no-download', '-d', default=CFG.get('download', True),
              help='Whether to download any existing translations prior'
                   ' to uploading.')
@click.option('--extract/--no-extract', '-x', default=CFG.get('extract', True),
              help='Whether to extract translations prior to uploading.')
@click.option('--prune', default=False, is_flag=True,
              help='Whether to remove obsolete messages from spreadsheet.'
                   ' Normally message may be removed and readded periodically'
                   ' so they are not removed, but can be removed to clean up'
                   ' translations but cannot be retrieved once pruned.')
@click.option('--update-meta', default=False, is_flag=True,
              help='Whether to update the meta information for the uploaded'
                   ' resources instead of uploading new translation files.'
                   ' This updates properties that are not directly related to'
                   ' the translation data, such as styling or filters')
@shared.locale_option(
    help_text='Which locale(s) to upload. If unspecified,'
              ' translations for all catalogs will be uploaded.')
@shared.service_option
def translations_upload(pod_path, locale, force, service, update_acl,
                        download, extract, prune, update_meta):
    """Uploads translations to a translation service."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    pod = pods.Pod(root, storage=storage.FileStorage)
    translator = pod.get_translator(service)

    if not translator:
        raise click.ClickException('No translators specified in podspec.yaml.')

    if update_acl:
        with pod.profile.timer('grow_translations_acl'):
            translator.update_acl(locales=locale)
        return

    if update_meta:
        with pod.profile.timer('grow_translations_meta'):
            translator.update_meta(locales=locale)
        return

    if download:
        with pod.profile.timer('grow_translations_download'):
            translator.download(locales=locale)

    if extract:
        with pod.profile.timer('grow_translations_extract'):
            include_obsolete, localized, include_header, use_fuzzy_matching, = \
                pod.catalogs.get_extract_config()
            catalogs = pod.get_catalogs()
            locales = validate_locales(pod.list_locales(), locale)
            catalogs.extract(include_obsolete=include_obsolete, localized=localized,
                             include_header=include_header,
                             use_fuzzy_matching=use_fuzzy_matching, locales=locales)
            if not localized:
                catalogs.update(locales=locales, include_header=include_header,
                                use_fuzzy_matching=use_fuzzy_matching)

    with pod.profile.timer('grow_translations_upload'):
        translator.upload(locales=locale, force=force,
                          verbose=True, prune=prune)

    if translator.needs_meta_update:
        with pod.profile.timer('grow_translations_meta'):
            translator.update_meta(locales=locale)

    return pod




"""Subcommand for deploying pod."""

import os
import click
from grow.commands import shared
from grow.common import bulk_errors
from grow.common import rc_config
from grow.common import utils
from grow.deployments import stats
from grow.deployments.destinations import base
from grow.extensions import hooks
from grow.performance import docs_loader
from grow.pods import pods
from grow.rendering import renderer
from grow import storage


CFG = rc_config.RC_CONFIG.prefixed('grow.deploy')


@click.command()
@click.argument('deployment_name', default='default')
@shared.pod_path_argument
@click.option('--confirm/--noconfirm', '-c/-f', default=CFG.get('force', True), is_flag=True,
              help='Whether to confirm prior to deployment.')
@click.option('--test/--notest', default=CFG.get('test', True), is_flag=True,
              help='Whether to run deployment tests.')
@click.option('--test_only', default=False, is_flag=True,
              help='Only run the deployment tests.')
@click.option('--auth',
              help='(deprecated) --auth must now be specified'
                   ' before deploy. Usage: grow --auth=user@example.com deploy')
@shared.force_untranslated_option(CFG)
@shared.preprocess_option(CFG)
@shared.threaded_option(CFG)
@shared.shards_option
@shared.shard_option
@shared.work_dir_option
@shared.routes_file_option()
@click.pass_context
def deploy(context, deployment_name, pod_path, preprocess, confirm, test,
           test_only, auth, force_untranslated, threaded, shards, shard,
           work_dir, routes_file):
    """Deploys a pod to a destination."""
    if auth:
        text = ('--auth must now be specified before deploy. Usage:'
                ' grow --auth=user@example.com deploy')
        raise click.ClickException(text)
    auth = context.parent.params.get('auth')
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    try:
        pod = pods.Pod(root, storage=storage.FileStorage)
        with pod.profile.timer('grow_deploy'):
            # Always clear the cache when building.
            pod.podcache.reset()
            deployment = pod.get_deployment(deployment_name)
            # use the deployment's environment for preprocessing and later
            # steps.
            if deployment.config.env:
                pod.set_env(deployment.config.env)
            localization = pod.podspec.localization
            require_translations = localization and localization.get(
                'require_translations', False)
            require_translations = require_translations and not force_untranslated
            if auth:
                deployment.login(auth)
            if preprocess:
                pod.preprocess()
            if test_only:
                deployment.test()
                return
            repo = utils.get_git_repo(pod.root)
            pod.router.use_simple()
            if routes_file:
                pod.router.from_data(pod.read_json(routes_file))
            else:
                pod.router.add_all()
            is_partial = False
            # Filter routes based on deployment config.
            for build_filter in deployment.filters:
                is_partial = True
                pod.router.filter(
                    build_filter.type, collection_paths=build_filter.collections,
                    paths=build_filter.paths, locales=build_filter.locales)

            # Shard the routes when using sharding.
            if shards and shard:
                is_partial = True
                pod.router.shard(shards, shard)

            if not work_dir:
                # Preload the documents used by the paths after filtering.
                docs_loader.DocsLoader.load_from_routes(pod, pod.router.routes)

            paths = pod.router.routes.paths
            stats_obj = stats.Stats(pod, paths=paths)
            content_generator = deployment.dump(
                pod, source_dir=work_dir, use_threading=threaded)
            content_generator = hooks.generator_wrapper(
                pod, 'pre_deploy', content_generator, 'deploy')
            deployment.deploy(
                content_generator, stats=stats_obj, repo=repo, confirm=confirm,
                test=test, require_translations=require_translations,
                is_partial=is_partial)
            pod.podcache.write()
    except bulk_errors.BulkErrors as err:
        # Write the podcache files even when there are rendering errors.
        pod.podcache.write()
        bulk_errors.display_bulk_errors(err)
        raise click.Abort()
    except base.Error as err:
        raise click.ClickException(str(err))
    except pods.Error as err:
        raise click.ClickException(str(err))
    return pod



"""Untag an object using convention based keys."""

import re
from boltons import iterutils


LOCALIZED_KEY_REGEX = re.compile(r'(.*)@([^@]+)$')


class Untag(object):
    """Untagging utility for locale and environment based untagging."""

    @staticmethod
    def untag(data, locale_identifier=None, params=None):
        """Untags fields, handling translation priority."""
        paths_to_keep_tagged = set()
        locale_identifier = str(locale_identifier)

        # When untagging the order of the keys isn't consistent. Sometimes the
        # tagged value is found but then is overwritten by the original value
        # since it is processed after the tagged version. Need to keep track of
        # untagged keys to make sure that they are not overwritten by the original.
        untagged_key_paths = set()

        # pylint: disable=too-many-return-statements
        def _visit(path, key, value):
            """Function for each key and value in the data."""
            if not isinstance(key, str):
                return key, value

            if key.endswith('@#'):
                # Translation Comment.
                return False

            marked_for_extraction = key.endswith('@')
            if marked_for_extraction:
                if isinstance(value, list):
                    paths_to_keep_tagged.add((path, key))
                key = key[:-1]

            # Support <key>@<param key>.<param value>: <value>.
            if params:
                param_regex = re.compile(
                    r'(.*)@({})\.([^@]+)$'.format('|'.join(params.keys())),
                    re.IGNORECASE)
                param_match = param_regex.match(key)
                if param_match:
                    untagged_key, param_key, param_value = param_match.groups()

                    # If the key has already been untagged, don't overwrite.
                    if (path, untagged_key) in untagged_key_paths:
                        return False

                    if not params[param_key]:
                        return False
                    result = params[param_key](
                        data, untagged_key, param_key, param_value, value,
                        locale_identifier=locale_identifier)
                    if result is not False:
                        # Don't let the original key overwrite the new value.
                        untagged_key_paths.add((path, untagged_key))
                    return result

            # Support <key>@<locale regex>: <value>.
            match = LOCALIZED_KEY_REGEX.match(key)
            if not match:
                if (path, key) in untagged_key_paths:
                    return False
                return key, value

            if not locale_identifier:
                return False

            untagged_key, locale_from_key = match.groups()

            # If the key has already been untagged, don't overwrite.
            if (path, untagged_key) in untagged_key_paths:
                return False

            # TODO: Once the translation process is able to correctly extract
            # the locale tagged extractions we need to prevent replacing.
            # # When marked for extraction when tagged it should be used as the
            # # translation value in the message catalog, not replace the value.
            # if marked_for_extraction:
            #     return False

            locale_regex = re.compile(
                r'^{}$'.format(locale_from_key), re.IGNORECASE)
            if not locale_regex.match(locale_identifier):
                return False

            # Don't let the original key overwrite the new value.
            untagged_key_paths.add((path, untagged_key))

            return untagged_key, value

        # Backwards compatibility for https://github.com/grow/grow/issues/95
        def _remap_exit(path, key, old_parent, new_parent, new_items):
            resp = iterutils.default_exit(path, key, old_parent,
                                          new_parent, new_items)
            if paths_to_keep_tagged and isinstance(resp, dict):
                updated_values = {}
                for sub_key, value in resp.items():
                    if not isinstance(value, list):
                        continue
                    new_key = '{}@'.format(sub_key)
                    updated_values[new_key] = value
                resp.update(updated_values)
                try:
                    paths_to_keep_tagged.remove((path, key))
                except KeyError:
                    pass
            return resp

        return iterutils.remap(data, visit=_visit, exit=_remap_exit)


class UntagParam(object):
    """Untagging param for complex untagging."""

    def __call__(self, data, untagged_key, param_key, param_value, value, locale_identifier=None):
        raise NotImplementedError()


class UntagParamRegex(object):
    """Param using the value of the param value as a regex to match."""

    def __init__(self, value):
        self.value = value

    def __call__(self, data, untagged_key, param_key, param_value, value, locale_identifier=None):
        if not self.value:
            return False
        value_regex = r'^{}$'.format(param_value)
        if not re.match(value_regex, self.value, re.IGNORECASE):
            return False
        return untagged_key, value


class UntagParamLocaleRegex(object):
    """Param using a document field as a regex group to match locale.

    Attempts to use the value of one of the other data fields as a locale regex.
    If there is no matching key found it falls back to the collection or podspec.
    """

    def __init__(self, podspec_data=None, collection_data=None):
        self.podspec_data = podspec_data or {}
        self.collection_data = collection_data or {}

    def __call__(self, data, untagged_key, param_key, param_value, value, locale_identifier=None):
        podspec_value = self.podspec_data.get(param_value, None)
        collection_value = self.collection_data.get(param_value, podspec_value)
        local_groups = data.get('$localization', {}).get('groups', {})
        regex_value = local_groups.get(param_value, collection_value)
        if not regex_value:
            return False

        # Convert lists of locales into a regex pattern.
        if not isinstance(regex_value, str):
            regex_value = '|'.join(regex_value)

        value_regex = r'^{}$'.format(regex_value)
        if not re.match(value_regex, locale_identifier, re.IGNORECASE):
            return False
        return untagged_key, value

    @classmethod
    def from_pod(cls, pod, collection=None):
        """Shortcut from the pod and collection objects."""
        podspec_data = pod.yaml.get('localization', {}).get('groups', {})
        if collection:
            collection_data = collection.fields.get('localization', {}).get('groups', {})
        else:
            collection_data = {}
        return cls(podspec_data, collection_data)



"""Base class for the pre render hook."""

from grow.extensions.hooks import base_hook


class PreRenderHook(base_hook.BaseHook):
    """Hook for pre render."""

    KEY = 'pre_render'
    NAME = 'Pre Render'

    # pylint: disable=arguments-differ,unused-argument
    def trigger(self, previous_result, doc, original_body, *_args, **_kwargs):
        """Trigger the pre render hook."""
        if previous_result:
            return previous_result
        # Return None if nothing has changed.
        return None



"""Local file storage."""

import errno
import hashlib
import os
import shutil
from grow.storage import base_storage
from grow.templates import jinja_loader


class FileStorage(base_storage.BaseStorage):

    @staticmethod
    def open(filename, mode=None):
        if mode is None:
            mode = 'r'
        return open(filename, mode=mode)

    @staticmethod
    def read(filename):
        try:
            with open(filename) as fp:
                content = fp.read()
        except UnicodeDecodeError as e:
            with open(filename, 'rb') as fp:
                content = fp.read()
        return content

    @staticmethod
    def modified(filename):
        return os.stat(filename).st_mtime

    @staticmethod
    def size(filename):
        return os.path.getsize(filename)

    @staticmethod
    def stat(filename):
        return os.stat(filename)

    @staticmethod
    def hash(filename):
        hash_digest = hashlib.sha256()
        with open(filename, "rb") as source_file:
            for chunk in iter(lambda: source_file.read(4096), b""):
                hash_digest.update(chunk)
        return hash_digest.hexdigest()

    @staticmethod
    def listdir(dirpath, recursive=True):
        paths = []
        for root, _, files in os.walk(dirpath, topdown=True, followlinks=True):
            for filename in files:
                path = os.path.join(root, filename)[len(dirpath):]
                paths.append(path)
            # if not recursive, break after walking top-level dir
            if not recursive:
                break
        return paths

    @staticmethod
    def walk(dirpath):
        return os.walk(dirpath, followlinks=True)

    @staticmethod
    def JinjaLoader(path):
        return jinja_loader.FrontMatterLoader(path)

    @classmethod
    def write(cls, path, content):
        dirname = os.path.dirname(path)
        try:
            os.makedirs(dirname)
        except OSError as e:
            if e.errno == errno.EEXIST and os.path.isdir(dirname):
                pass
            else:
                raise
        try:
            with cls.open(path, mode='w') as fp:
                fp.write(content)
        except TypeError:
            with cls.open(path, mode='wb') as fp:
                fp.write(content)

    @staticmethod
    def exists(filename):
        return os.path.exists(filename)

    @staticmethod
    def delete(filename):
        return os.remove(filename)

    @staticmethod
    def delete_dir(dirpath):
        shutil.rmtree(dirpath)

    @staticmethod
    def delete_files(dirpaths, recursive=False, pattern=None):
        """Delete files from within the dirpaths that match a pattern."""
        for dirpath in dirpaths:
            for root, _, files in os.walk(dirpath, followlinks=True):
                for filename in files:
                    if not pattern or pattern.search(filename):
                        os.remove(os.path.join(root, filename))
                if not recursive:
                    break

    @staticmethod
    def copy_to(paths, target_paths):
        # TODO(jeremydw): Rename to bulk_copy_to.
        for i, path in enumerate(paths):
            target_path = target_paths[i]
            shutil.copyfile(path, target_path)
            shutil.copystat(path, target_path)

    @staticmethod
    def move_to(path, target_path):
        os.rename(path, target_path)



"""Utility class for tracking a doc dependency."""


# pylint: disable=too-few-public-methods
class DocDependency(object):
    """Used to easily track a dependencies for a document."""

    def __init__(self, doc):
        self.doc = doc
        self.dependency_graph = self.doc.pod.podcache.dependency_graph if doc else None

    def __call__(self, pod_path):
        if self.dependency_graph and pod_path:
            self.dependency_graph.add(self.doc.pod_path, pod_path)



"""Caching for pod meta information."""

import json
import os
import tempfile
from grow.cache import collection_cache
from grow.cache import document_cache
from grow.cache import file_cache
from grow.cache import object_cache
from grow.cache import routes_cache as grow_routes_cache
from grow.common import json_encoder
from grow.pods import dependency


FILE_OBJECT_CACHE = object_cache.FILE_OBJECT_CACHE
FILE_ROUTES_CACHE = grow_routes_cache.FILE_ROUTES_CACHE


class Error(Exception):
    """General podcache error."""

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class PodCacheParseError(Error):
    """Error parsing the podcache file."""
    pass


class PodCache(object):
    """Caching container for the pod."""

    KEY_GLOBAL = '__global__'

    def __init__(self, dep_cache, obj_cache, routes_cache, pod):
        self._pod = pod

        self._collection_cache = collection_cache.CollectionCache()
        self._document_cache = document_cache.DocumentCache()
        self._file_cache = file_cache.FileCache()

        self._dependency_graph = dependency.DependencyGraph()
        self._dependency_graph.add_all(dep_cache)

        self._object_caches = {}
        self.create_object_cache(
            self.KEY_GLOBAL, write_to_file=False, can_reset=True)

        for key, item in obj_cache.items():
            # If this is a string, it is written to a separate cache file.
            if isinstance(item, str):
                cache_value = {}
                if self._pod.file_exists(item):
                    cache_value = self._pod.read_json(item)
                self.create_object_cache(key, **cache_value)
            else:
                self.create_object_cache(key, **item)

        self._routes_cache = grow_routes_cache.RoutesCache()
        self._routes_cache.from_data(routes_cache)

    @property
    def collection_cache(self):
        """Cache for the collections."""
        return self._collection_cache

    @property
    def dependency_graph(self):
        """Dependency graph from rendered docs."""
        return self._dependency_graph

    @property
    def document_cache(self):
        """Cache for specific document properties."""
        return self._document_cache

    @property
    def file_cache(self):
        """Cache for raw file contents."""
        return self._file_cache

    @property
    def is_dirty(self):
        """Have the contents of the dependency graph or caches been modified?"""
        if self.dependency_graph.is_dirty:
            return True
        for meta in self._object_caches.values():
            if meta['write_to_file'] and meta['cache'].is_dirty:
                return True
        if self.routes_cache.is_dirty:
            return True
        return False

    @property
    def object_cache(self):
        """Global object cache."""
        return self.get_object_cache(self.KEY_GLOBAL)

    @property
    def routes_cache(self):
        """Global routes cache."""
        return self._routes_cache

    def _write_json(self, path, obj):
        output = json.dumps(
            obj, cls=json_encoder.GrowJSONEncoder,
            sort_keys=True, indent=2, separators=(',', ': '))
        temp_dir = self._pod.abs_path('/.grow/tmp')
        if not os.path.exists(temp_dir):
            os.makedirs(temp_dir)
        fd, temp_path = tempfile.mkstemp(dir=temp_dir)
        with os.fdopen(fd, 'w') as fp:
            fp.write(output)
        real_path = self._pod.abs_path(path)
        dir_name = os.path.dirname(real_path)
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        os.rename(temp_path, real_path)

    def create_object_cache(self, key, write_to_file=False, can_reset=False, values=None,
                            separate_file=False):
        """Create a named object cache."""
        self._object_caches[key] = {
            'cache': object_cache.ObjectCache(),
            'write_to_file': write_to_file,
            'can_reset': can_reset,
            'separate_file': separate_file,
        }
        cache = self._object_caches[key]['cache']
        if values:
            cache.add_all(values)
        return cache

    def get_object_cache(self, key, **kwargs):
        """Get an existing object cache or create a new cache with defaults."""
        if key not in self._object_caches:
            return self.create_object_cache(key, **kwargs)
        existing_meta = self._object_caches[key]
        if 'separate_file' in kwargs:
            if existing_meta['separate_file'] != kwargs['separate_file']:
                existing_meta['separate_file'] = kwargs['separate_file']
        return existing_meta['cache']

    def has_object_cache(self, key):
        """Has an existing object cache?"""
        return key in self._object_caches

    def reset(self, force=False):
        """Reset pod caches."""
        self._collection_cache.reset()
        self._dependency_graph.reset()
        self._document_cache.reset()
        self._file_cache.reset()

        # Only reset the object caches if permitted.
        for meta in self._object_caches.values():
            if meta['can_reset'] or force:
                meta['cache'].reset()

    def update(self, dep_cache=None, obj_cache=None):
        """Update the values in the dependency cache and/or the object cache."""
        if dep_cache:
            self._dependency_graph.add_all(dep_cache)

        if obj_cache:
            for key, meta in obj_cache.items():
                if not key in self._object_caches:
                    self.create_object_cache(key, **meta)
                else:
                    # Ignore if the object cache is referenced to a different file.
                    if isinstance(meta, str):
                        continue

                    self._object_caches[key]['cache'].add_all(meta['values'])
                    self._object_caches[key][
                        'write_to_file'] = meta['write_to_file']
                    self._object_caches[key][
                        'separate_file'] = meta['separate_file']
                    self._object_caches[key]['can_reset'] = meta['can_reset']

    def write(self):
        """Persist the cache information to a yaml file."""
        with self._pod.profile.timer('Podcache.write'):
            if self._dependency_graph.is_dirty:
                output = self._dependency_graph.export()
                self._dependency_graph.mark_clean()
                self._write_json('{}{}'.format(
                    self._pod.PATH_CONTROL, self._pod.FILE_DEP_CACHE), output)

            if self._routes_cache.is_dirty:
                output = self._routes_cache.export()
                self._routes_cache.mark_clean()
                self._write_json('{}{}'.format(
                    self._pod.PATH_CONTROL, FILE_ROUTES_CACHE), output)

            # Write out any of the object caches configured for write_to_file.
            output = {}
            for key, meta in self._object_caches.items():
                if meta['write_to_file']:
                    cache_info = {
                        'can_reset': meta['can_reset'],
                        'write_to_file': meta['write_to_file'],
                        'values': meta['cache'].export(),
                        'separate_file': meta['separate_file'],
                    }
                    if meta['separate_file']:
                        filename = '/{}'.format(
                            object_cache.FILE_OBJECT_SUB_CACHE.format(key))
                        self._write_json(filename, cache_info)
                        output[key] = filename
                    else:
                        output[key] = cache_info
                    meta['cache'].mark_clean()
            if output:
                self._write_json('/{}'.format(FILE_OBJECT_CACHE), output)



"""Command to run the local development server."""
import os
import click
from grow.commands import shared
from grow.common import rc_config
from grow.pods import env
from grow.pods import pods
from grow import storage
from grow.server import manager


CFG = rc_config.RC_CONFIG.prefixed('grow.run')


# pylint: disable=too-many-locals, invalid-name, too-many-arguments
@click.command()
@shared.pod_path_argument
@click.option('--host', default=CFG.get('host', 'localhost'))
@click.option('--port', default=CFG.get('port', 8080))
@click.option('--https', default=CFG.get('https', False),
              help='Whether to use "https" in the local environment.')
@click.option('--debug/--no-debug', default=CFG.get('debug', False), is_flag=True,
              help='Whether to run in debug mode and show internal tracebacks'
                   ' when encountering exceptions.')
@click.option('--browser/--no-browser', '-b', is_flag=True,
              default=CFG.get('browser', False),
              help='Whether to open a browser upon startup.')
@click.option('--update-check/--no-update-check',
              default=CFG.get('update-check', True), is_flag=True,
              help='Whether to check for updates to Grow.')
@click.option('--ui/--no-ui', is_flag=CFG.get('ui', True), default=True,
              help='Whether to inject the Grow UI Tools.')
@shared.deployment_option(CFG)
@shared.preprocess_option(CFG)
def run(host, port, https, debug, browser, update_check, preprocess, ui,
        pod_path, deployment):
    """Starts a development server for a single pod."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    scheme = 'https' if https else 'http'
    config = env.EnvConfig(host=host, port=port, name=env.Name.DEV,
                           scheme=scheme, cached=False, dev=True)
    environment = env.Env(config)
    pod = pods.Pod(
        root, storage=storage.FileStorage, env=environment)
    if deployment:
        deployment_obj = pod.get_deployment(deployment)
        pod.set_env(deployment_obj.config.env)
    if not ui:
        pod.disable(pod.FEATURE_UI)
    try:
        manager.start(pod, host=host, port=port, open_browser=browser,
                      debug=debug, preprocess=preprocess,
                      update_check=update_check)
    except pods.Error as e:
        raise click.ClickException(str(e))
    except KeyboardInterrupt:
        pass  # Let the normal process work when doing keyboard inturrupt.
    return pod
